#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server.contexts import datasets
from argilla_server.database import get_async_db
from argilla_server.models import Suggestion, User
from argilla_server.policies import SuggestionPolicyV1, authorize
from argilla_server.schemas.v1.suggestions import Suggestion as SuggestionSchema
from argilla_server.search_engine import SearchEngine, get_search_engine
from argilla_server.security import auth

router = APIRouter(tags=["suggestions"])


async def _get_suggestion(db: "AsyncSession", suggestion_id: UUID) -> Suggestion:
    suggestion = await datasets.get_suggestion_by_id(db, suggestion_id)
    if not suggestion:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Suggestion with id `{suggestion_id}` not found",
        )
    return suggestion


@router.delete("/suggestions/{suggestion_id}", response_model=SuggestionSchema)
async def delete_suggestion(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    suggestion_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    suggestion = await _get_suggestion(db, suggestion_id)

    await authorize(current_user, SuggestionPolicyV1.delete(suggestion))

    try:
        return await datasets.delete_suggestion(db, search_engine, suggestion)
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, List, Optional

from argilla_server.commons.models import PredictionStatus, TaskType
from argilla_server.pydantic_v1 import BaseModel, Field
from argilla_server.services.search.model import ServiceBaseRecordsQuery, ServiceScoreRange
from argilla_server.services.tasks.commons import ServiceBaseAnnotation, ServiceBaseRecord


class ServiceText2TextPrediction(BaseModel):
    text: str
    score: float


class ServiceText2TextAnnotation(ServiceBaseAnnotation):
    sentences: List[ServiceText2TextPrediction]


class ServiceText2TextRecord(ServiceBaseRecord[ServiceText2TextAnnotation]):
    text: str

    @classmethod
    def task(cls) -> TaskType:
        """The task type"""
        return TaskType.text2text

    def all_text(self) -> str:
        return self.text

    @property
    def predicted_as(self) -> Optional[List[str]]:
        return [sentence.text for sentence in self.prediction.sentences] if self.prediction else None

    @property
    def annotated_as(self) -> Optional[List[str]]:
        return [sentence.text for sentence in self.annotation.sentences] if self.annotation else None

    @property
    def scores(self) -> List[float]:
        """Values of prediction scores"""
        if not self.prediction:
            return []
        return [sentence.score for sentence in self.prediction.sentences]

    def extended_fields(self) -> Dict[str, Any]:
        return {
            "annotated_as": self.annotated_as,
            "predicted_as": self.predicted_as,
            "annotated_by": self.annotated_by,
            "predicted_by": self.predicted_by,
            "score": self.scores,
        }


class ServiceText2TextQuery(ServiceBaseRecordsQuery):
    score: Optional[ServiceScoreRange] = Field(default=None)
    predicted: Optional[PredictionStatus] = Field(default=None, nullable=True)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, Optional, Union

from fastapi import APIRouter, Depends, Security

from argilla_server.apis.v0.models.commons.params import CommonTaskHandlerDependencies
from argilla_server.apis.v0.models.text2text import Text2TextRecord
from argilla_server.apis.v0.models.text_classification import TextClassificationRecord
from argilla_server.apis.v0.models.token_classification import TokenClassificationRecord
from argilla_server.commons.config import TasksFactory
from argilla_server.commons.models import TaskStatus
from argilla_server.models import User
from argilla_server.pydantic_v1 import BaseModel
from argilla_server.security import auth
from argilla_server.services.datasets import DatasetsService
from argilla_server.services.search.service import SearchRecordsService
from argilla_server.services.storage.service import RecordsStorageService


def configure_router(router: APIRouter):
    RecordType = Union[
        TextClassificationRecord,
        TokenClassificationRecord,
        Text2TextRecord,
    ]

    class PartialUpdateRequest(BaseModel):
        metadata: Optional[Dict[str, Any]] = None
        status: Optional[TaskStatus] = None

    @router.patch(
        "/{name}/records/{record_id}",
        response_model=RecordType,
    )
    async def partial_update_dataset_record(
        name: str,
        record_id: str,
        request: PartialUpdateRequest,
        request_deps: CommonTaskHandlerDependencies = Depends(),
        service: DatasetsService = Depends(DatasetsService.get_instance),
        search: SearchRecordsService = Depends(SearchRecordsService.get_instance),
        storage: RecordsStorageService = Depends(RecordsStorageService.get_instance),
        current_user: User = Security(auth.get_current_user, scopes=[]),
    ) -> RecordType:
        dataset = await service.find_by_name(
            user=current_user,
            name=name,
            workspace=request_deps.workspace,
        )

        record_class = TasksFactory.get_task_record(dataset.task)
        record = await search.find_record_by_id(
            dataset=dataset,
            id=record_id,
            record_type=record_class,
        )

        return await storage.update_record(
            dataset=dataset,
            record=record,
            metadata=request.metadata,
            status=request.status,
        )


router = APIRouter(tags=["records"], prefix="/datasets")
configure_router(router)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from fastapi import APIRouter, Depends, HTTPException, Request
from fastapi.responses import RedirectResponse
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server import telemetry
from argilla_server.contexts import accounts
from argilla_server.database import get_async_db
from argilla_server.enums import UserRole
from argilla_server.errors.future import AuthenticationError
from argilla_server.models import User
from argilla_server.schemas.v1.oauth2 import Provider, Providers, Token
from argilla_server.security.authentication.jwt import JWT
from argilla_server.security.authentication.oauth2 import OAuth2ClientProvider
from argilla_server.security.authentication.userinfo import UserInfo
from argilla_server.security.settings import settings

router = APIRouter(prefix="/oauth2", tags=["Authentication"])


_USER_ROLE_ON_CREATION = UserRole.annotator


@router.get("/providers", response_model=Providers)
def list_providers(_request: Request) -> Providers:
    if not settings.oauth.enabled:
        return Providers(items=[])

    return Providers(items=[Provider(name=provider_name) for provider_name in settings.oauth.providers])


@router.get("/providers/{provider}/authentication")
def get_authentication(request: Request, provider: str) -> RedirectResponse:
    _check_oauth_enabled_or_raise()

    provider = _get_provider_by_name_or_raise(provider)
    return provider.authorization_redirect(request)


@router.get("/providers/{provider}/access-token", response_model=Token)
async def get_access_token(
    request: Request,
    provider: str,
    db: AsyncSession = Depends(get_async_db),
) -> Token:
    _check_oauth_enabled_or_raise()

    try:
        provider = _get_provider_by_name_or_raise(provider)
        user_info = UserInfo(await provider.get_user_data(request))

        user_info.use_claims(provider.claims)
        username = user_info.username

        user = await accounts.get_user_by_username(db, username)
        if user is None:
            user = await accounts.create_user_with_random_password(
                db,
                username=username,
                first_name=user_info.name,
                role=_USER_ROLE_ON_CREATION,
                workspaces=[workspace.name for workspace in settings.oauth.allowed_workspaces],
            )
            telemetry.track_user_created(user, is_oauth=True)
        elif not _is_user_created_by_oauth_provider(user):
            # User should sign in using username/password workflow
            raise AuthenticationError("Could not authenticate user")

        return Token(access_token=JWT.create(user_info))
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))
    except AuthenticationError as e:
        raise HTTPException(status_code=401, detail=str(e))


def _check_oauth_enabled_or_raise() -> None:
    if not settings.oauth.enabled:
        raise HTTPException(status_code=404, detail="OAuth2 is not enabled")


def _get_provider_by_name_or_raise(provider_name: str) -> OAuth2ClientProvider:
    if not provider_name in settings.oauth.providers:
        raise HTTPException(status_code=404, detail=f"Provider '{provider_name}' not found")
    return settings.oauth.providers[provider_name]


def _is_user_created_by_oauth_provider(user: User) -> bool:
    # TODO: We must link the created user with the provider, and base this check on that.
    #  For now, we just validate the user role on creation.
    return user.role == _USER_ROLE_ON_CREATION



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""add allow_extra_metadata column to dataset table

Revision ID: b8458008b60e
Revises: 7cbcccf8b57a
Create Date: 2023-09-29 13:51:44.525944

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b8458008b60e"
down_revision = "7cbcccf8b57a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "datasets", sa.Column("allow_extra_metadata", sa.Boolean(), server_default=sa.text("true"), nullable=False)
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("datasets", "allow_extra_metadata")
    # ### end Alembic commands ###



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import importlib
import sys
from typing import Any, List

import pytest
from argilla_server.utils.dependency import (
    is_package_with_extras_installed,
    require_dependencies,
    requires_dependencies,
)


class TestDependencyRequirements:
    @pytest.mark.parametrize(
        ("decorator", "package_name", "import_name", "version"),
        [
            (requires_dependencies("datasets>1.17.0"), "datasets", "datasets", ">1.17.0"),
            (requires_dependencies("spacy"), "spacy", "spacy", ""),
            (requires_dependencies("scikit-learn"), "scikit-learn", "sklearn", ""),
            (requires_dependencies("faiss"), "faiss", "faiss", ""),
        ],
    )
    def test_missing_dependency_decorator(
        self, monkeypatch: pytest.MonkeyPatch, decorator, package_name: str, import_name: str, version: str
    ):
        monkeypatch.setitem(sys.modules, import_name, None)
        monkeypatch.setattr(sys, "meta_path", [], raising=False)

        # Ensure that the package indeed cannot be imported due to the monkeypatch
        with pytest.raises(ModuleNotFoundError):
            importlib.import_module(import_name)

        @decorator
        def test_inner():
            pass

        requirement = package_name + version
        # Verify that the decorator does its work and shows the desired output with `pip install ...`
        with pytest.raises(
            ModuleNotFoundError,
            match=f"'{package_name}' must be installed to use `test_inner`!.*?`pip install {requirement}`",
        ):
            test_inner()

    @pytest.mark.parametrize(
        ("decorator"),
        [
            requires_dependencies("datasets>1.17.0"),
            requires_dependencies("spacy"),
            requires_dependencies("scikit-learn"),
            requires_dependencies(["datasets>1.17.0", "spacy", "scikit-learn"]),
        ],
    )
    def test_installed_dependency_decorator(self, decorator):
        # Ensure that the decorated function can be called just fine if the dependencies are installed,
        # which they should be for these tests

        @decorator
        def test_inner():
            return True

        assert test_inner()

    def test_installed_dependency_but_incorrect_version(self):
        def test_inner():
            require_dependencies("datasets<1.0.0")
            return True

        # This method should fail, as our dependencies require a higher version of datasets
        with pytest.raises(
            ImportError,
            match=f"but found datasets==.*?You can install a supported version of 'datasets' with this command: `pip install -U datasets<1.0.0`",
        ):
            test_inner()

    def test_require_version_failures(self):
        # This operation is not supported
        with pytest.raises(ValueError):
            require_dependencies("datasets~=1.0.0")

        # Add some unsupported tokens, e.g. " "
        with pytest.raises(ValueError):
            require_dependencies(" datasets")

        # Add unsupported operation in second requirement version
        with pytest.raises(ValueError):
            require_dependencies("datasets>1.0.0,~1.17.0")

    def test_special_python_case(self):
        require_dependencies("python>3.6")

    def test_multiple_version_requirements(self):
        # This is equivalent to just datasets>1.17.0, but we expect it to work still
        require_dependencies("datasets>1.0.0,>1.8.0,>1.17.0")
        # A more common example (designed not to break eventually):
        require_dependencies("datasets>1.17.0,<1000.0.0")

    def test_list_of_dependencies(self):
        require_dependencies(["datasets>1.17.0", "spacy", "scikit-learn"])

    def test_list_without_dependencies(self):
        with pytest.raises(ValueError, match="requirements cannot be an empty list."):
            require_dependencies([])


@pytest.mark.skip(reason="Not working as expected with new argilla_server package")
@pytest.mark.parametrize(
    "args, expected", [(["argilla", ["server"]], True), (["invented_package", ["invented"]], False)]
)
def test_is_package_with_extras_installed(args: List[Any], expected: bool) -> None:
    assert is_package_with_extras_installed(*args) == expected


@pytest.mark.skip(reason="Not working as expected with new argilla_server package")
def test_is_package_with_extras_installed_raises_key_error() -> None:
    with pytest.raises(KeyError, match="'argilla' package does not provide 'invented' extra"):
        is_package_with_extras_installed("argilla", ["invented"])



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterable, List, Optional

from fastapi import Depends

from argilla_server.commons.config import TasksFactory
from argilla_server.services.datasets import ServiceDataset
from argilla_server.services.search.model import ServiceSearchResults, ServiceSortableField, ServiceSortConfig
from argilla_server.services.search.service import SearchRecordsService
from argilla_server.services.storage.service import RecordsStorageService
from argilla_server.services.tasks.commons import BulkResponse
from argilla_server.services.tasks.text2text.models import ServiceText2TextQuery, ServiceText2TextRecord


class Text2TextService:
    """
    Text2text service

    """

    _INSTANCE: "Text2TextService" = None

    @classmethod
    def get_instance(
        cls,
        storage: RecordsStorageService = Depends(RecordsStorageService.get_instance),
        search: SearchRecordsService = Depends(SearchRecordsService.get_instance),
    ):
        if not cls._INSTANCE:
            cls._INSTANCE = cls(storage, search)
        return cls._INSTANCE

    def __init__(
        self,
        storage: RecordsStorageService,
        search: SearchRecordsService,
    ):
        self.__storage__ = storage
        self.__search__ = search

    async def add_records(
        self,
        dataset: ServiceDataset,
        records: List[ServiceText2TextRecord],
    ):
        failed = await self.__storage__.store_records(
            dataset=dataset,
            records=records,
            record_type=ServiceText2TextRecord,
        )
        return BulkResponse(dataset=dataset.name, processed=len(records), failed=failed)

    def search(
        self,
        dataset: ServiceDataset,
        query: ServiceText2TextQuery,
        sort_by: List[ServiceSortableField],
        record_from: int = 0,
        size: int = 100,
        exclude_metrics: bool = True,
    ) -> ServiceSearchResults:
        metrics = TasksFactory.find_task_metrics(
            dataset.task,
            metric_ids={
                "words_cloud",
                "predicted_by",
                "annotated_by",
                "status_distribution",
                "metadata",
                "score",
            },
        )

        results = self.__search__.search(
            dataset,
            query=query,
            size=size,
            record_from=record_from,
            record_type=ServiceText2TextRecord,
            sort_config=ServiceSortConfig(
                sort_by=sort_by,
            ),
            exclude_metrics=exclude_metrics,
            metrics=metrics,
        )

        if results.metrics:
            results.metrics["words"] = results.metrics.get("words_cloud", {})
            results.metrics["status"] = results.metrics["status_distribution"]

        return results

    def read_dataset(
        self,
        dataset: ServiceDataset,
        query: Optional[ServiceText2TextQuery] = None,
        id_from: Optional[str] = None,
        limit: int = 1000,
    ) -> Iterable[ServiceText2TextRecord]:
        """
        Scan a dataset records

        Parameters
        ----------
        dataset:
            The records dataset
        query:
            If provided, scan will retrieve only records matching
            the provided query filters. Optional
        id_from:
            If provided, read the samples after this record ID
        limit:
            Batch size to scan, only used if `id_from` is specified

        """
        yield from self.__search__.scan_records(
            dataset,
            query=query,
            record_type=ServiceText2TextRecord,
            id_from=id_from,
            limit=limit,
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server.contexts import accounts, datasets
from argilla_server.database import get_async_db
from argilla_server.enums import ResponseStatusFilter
from argilla_server.models import Dataset as DatasetModel
from argilla_server.models import ResponseStatus, User
from argilla_server.policies import DatasetPolicyV1, MetadataPropertyPolicyV1, authorize, is_authorized
from argilla_server.schemas.v1.datasets import (
    Dataset,
    DatasetCreate,
    DatasetMetrics,
    Datasets,
    DatasetUpdate,
)
from argilla_server.schemas.v1.fields import Field, FieldCreate, Fields
from argilla_server.schemas.v1.metadata_properties import MetadataProperties, MetadataProperty, MetadataPropertyCreate
from argilla_server.schemas.v1.questions import Question, QuestionCreate, Questions
from argilla_server.schemas.v1.vector_settings import VectorSettings, VectorSettingsCreate, VectorsSettings
from argilla_server.search_engine import (
    SearchEngine,
    get_search_engine,
)
from argilla_server.security import auth
from argilla_server.telemetry import TelemetryClient, get_telemetry_client

CREATE_DATASET_VECTOR_SETTINGS_MAX_COUNT = 5

router = APIRouter()


async def _get_dataset(
    db: AsyncSession,
    dataset_id: UUID,
    with_fields: bool = False,
    with_questions: bool = False,
    with_metadata_properties: bool = False,
    with_vectors_settings: bool = False,
) -> DatasetModel:
    dataset = await datasets.get_dataset_by_id(
        db,
        dataset_id,
        with_fields=with_fields,
        with_questions=with_questions,
        with_metadata_properties=with_metadata_properties,
        with_vectors_settings=with_vectors_settings,
    )
    if not dataset:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Dataset with id `{dataset_id}` not found",
        )
    return dataset


async def _filter_metadata_properties_by_policy(
    current_user: User, metadata_properties: List[MetadataProperty]
) -> List[MetadataProperty]:
    filtered_metadata_properties = []

    for metadata_property in metadata_properties:
        metadata_property_is_authorized = await is_authorized(
            current_user, MetadataPropertyPolicyV1.get(metadata_property)
        )

        if metadata_property_is_authorized:
            filtered_metadata_properties.append(metadata_property)

    return filtered_metadata_properties


@router.get("/me/datasets", response_model=Datasets)
async def list_current_user_datasets(
    *,
    db: AsyncSession = Depends(get_async_db),
    workspace_id: Optional[UUID] = None,
    current_user: User = Security(auth.get_current_user),
):
    await authorize(current_user, DatasetPolicyV1.list(workspace_id))

    if not workspace_id:
        if current_user.is_owner:
            dataset_list = await datasets.list_datasets(db)
        else:
            await current_user.awaitable_attrs.datasets
            dataset_list = current_user.datasets
    else:
        dataset_list = await datasets.list_datasets_by_workspace_id(db, workspace_id)

    return Datasets(items=dataset_list)


@router.get("/datasets/{dataset_id}/fields", response_model=Fields)
async def list_dataset_fields(
    *, db: AsyncSession = Depends(get_async_db), dataset_id: UUID, current_user: User = Security(auth.get_current_user)
):
    dataset = await _get_dataset(db, dataset_id, with_fields=True)

    await authorize(current_user, DatasetPolicyV1.get(dataset))

    return Fields(items=dataset.fields)


@router.get("/datasets/{dataset_id}/questions", response_model=Questions)
async def list_dataset_questions(
    *, db: AsyncSession = Depends(get_async_db), dataset_id: UUID, current_user: User = Security(auth.get_current_user)
):
    dataset = await _get_dataset(db, dataset_id, with_questions=True)

    await authorize(current_user, DatasetPolicyV1.get(dataset))

    return Questions(items=dataset.questions)


@router.get("/datasets/{dataset_id}/vectors-settings", response_model=VectorsSettings)
async def list_dataset_vector_settings(
    *, db: AsyncSession = Depends(get_async_db), dataset_id: UUID, current_user: User = Security(auth.get_current_user)
):
    dataset = await _get_dataset(db, dataset_id, with_vectors_settings=True)

    await authorize(current_user, DatasetPolicyV1.get(dataset))

    return VectorsSettings(items=dataset.vectors_settings)


@router.get("/me/datasets/{dataset_id}/metadata-properties", response_model=MetadataProperties)
async def list_current_user_dataset_metadata_properties(
    *, db: AsyncSession = Depends(get_async_db), dataset_id: UUID, current_user: User = Security(auth.get_current_user)
):
    dataset = await _get_dataset(db, dataset_id, with_metadata_properties=True)

    await authorize(current_user, DatasetPolicyV1.get(dataset))

    filtered_metadata_properties = await _filter_metadata_properties_by_policy(
        current_user, dataset.metadata_properties
    )

    return MetadataProperties(items=filtered_metadata_properties)


@router.get("/datasets/{dataset_id}", response_model=Dataset)
async def get_dataset(
    *, db: AsyncSession = Depends(get_async_db), dataset_id: UUID, current_user: User = Security(auth.get_current_user)
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.get(dataset))

    return dataset


@router.get("/me/datasets/{dataset_id}/metrics", response_model=DatasetMetrics)
async def get_current_user_dataset_metrics(
    *, db: AsyncSession = Depends(get_async_db), dataset_id: UUID, current_user: User = Security(auth.get_current_user)
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.get(dataset))

    return {
        "records": {
            "count": await datasets.count_records_by_dataset_id(db, dataset_id),
        },
        "responses": {
            "count": await datasets.count_responses_by_dataset_id_and_user_id(db, dataset_id, current_user.id),
            "submitted": await datasets.count_responses_by_dataset_id_and_user_id(
                db, dataset_id, current_user.id, ResponseStatus(ResponseStatusFilter.submitted)
            ),
            "discarded": await datasets.count_responses_by_dataset_id_and_user_id(
                db, dataset_id, current_user.id, ResponseStatus(ResponseStatusFilter.discarded)
            ),
            "draft": await datasets.count_responses_by_dataset_id_and_user_id(
                db, dataset_id, current_user.id, ResponseStatus(ResponseStatusFilter.draft)
            ),
        },
    }


@router.post("/datasets", status_code=status.HTTP_201_CREATED, response_model=Dataset)
async def create_dataset(
    *,
    db: AsyncSession = Depends(get_async_db),
    dataset_create: DatasetCreate,
    current_user: User = Security(auth.get_current_user),
):
    await authorize(current_user, DatasetPolicyV1.create(dataset_create.workspace_id))

    if not await accounts.get_workspace_by_id(db, dataset_create.workspace_id):
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Workspace with id `{dataset_create.workspace_id}` not found",
        )

    if await datasets.get_dataset_by_name_and_workspace_id(db, dataset_create.name, dataset_create.workspace_id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Dataset with name `{dataset_create.name}` already exists for workspace with id `{dataset_create.workspace_id}`",
        )

    dataset = await datasets.create_dataset(db, dataset_create)
    return dataset


@router.post("/datasets/{dataset_id}/fields", status_code=status.HTTP_201_CREATED, response_model=Field)
async def create_dataset_field(
    *,
    db: AsyncSession = Depends(get_async_db),
    dataset_id: UUID,
    field_create: FieldCreate,
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.create_field(dataset))

    if await datasets.get_field_by_name_and_dataset_id(db, field_create.name, dataset_id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Field with name `{field_create.name}` already exists for dataset with id `{dataset_id}`",
        )

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        field = await datasets.create_field(db, dataset, field_create)
        return field
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.post("/datasets/{dataset_id}/questions", status_code=status.HTTP_201_CREATED, response_model=Question)
async def create_dataset_question(
    *,
    db: AsyncSession = Depends(get_async_db),
    dataset_id: UUID,
    question_create: QuestionCreate,
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.create_question(dataset))

    if await datasets.get_question_by_name_and_dataset_id(db, question_create.name, dataset_id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Question with name `{question_create.name}` already exists for dataset with id `{dataset_id}`",
        )

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        question = await datasets.create_question(db, dataset, question_create)
        return question
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.post(
    "/datasets/{dataset_id}/metadata-properties", status_code=status.HTTP_201_CREATED, response_model=MetadataProperty
)
async def create_dataset_metadata_property(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    dataset_id: UUID,
    metadata_property_create: MetadataPropertyCreate,
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.create_metadata_property(dataset))

    if await datasets.get_metadata_property_by_name_and_dataset_id(db, metadata_property_create.name, dataset_id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Metadata property with name `{metadata_property_create.name}` "
            f"already exists for dataset with id `{dataset_id}`",
        )

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        metadata_property = await datasets.create_metadata_property(
            db, search_engine, dataset, metadata_property_create
        )
        return metadata_property
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.post(
    "/datasets/{dataset_id}/vectors-settings", status_code=status.HTTP_201_CREATED, response_model=VectorSettings
)
async def create_dataset_vector_settings(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    dataset_id: UUID,
    vector_settings_create: VectorSettingsCreate,
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.create_vector_settings(dataset))

    count_vectors_settings_by_dataset_id = await datasets.count_vectors_settings_by_dataset_id(db, dataset_id)
    if count_vectors_settings_by_dataset_id >= CREATE_DATASET_VECTOR_SETTINGS_MAX_COUNT:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"The maximum number of vector settings has been reached for dataset with id `{dataset_id}`",
        )

    if await datasets.get_vector_settings_by_name_and_dataset_id(db, vector_settings_create.name, dataset_id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Vector settings with name `{vector_settings_create.name}` already exists for dataset with id"
            f" `{dataset_id}`",
        )

    try:
        vector_settings = await datasets.create_vector_settings(
            db, search_engine, dataset=dataset, vector_settings_create=vector_settings_create
        )
        return vector_settings
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.put("/datasets/{dataset_id}/publish", response_model=Dataset)
async def publish_dataset(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    telemetry_client: TelemetryClient = Depends(get_telemetry_client),
    dataset_id: UUID,
    current_user: User = Security(auth.get_current_user),
) -> DatasetModel:
    dataset = await _get_dataset(
        db, dataset_id, with_fields=True, with_questions=True, with_metadata_properties=True, with_vectors_settings=True
    )

    await authorize(current_user, DatasetPolicyV1.publish(dataset))
    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    #  After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        dataset = await datasets.publish_dataset(db, search_engine, dataset)

        telemetry_client.track_data(
            action="PublishedDataset",
            data={"questions": list(set([question.settings["type"] for question in dataset.questions]))},
        )

        return dataset
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.delete("/datasets/{dataset_id}", response_model=Dataset)
async def delete_dataset(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    dataset_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.delete(dataset))

    await datasets.delete_dataset(db, search_engine, dataset=dataset)

    return dataset


@router.patch("/datasets/{dataset_id}", response_model=Dataset)
async def update_dataset(
    *,
    db: AsyncSession = Depends(get_async_db),
    dataset_id: UUID,
    dataset_update: DatasetUpdate,
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.update(dataset))

    return await datasets.update_dataset(db, dataset=dataset, dataset_update=dataset_update)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import TYPE_CHECKING, List, Optional

import pytest
from argilla_server.apis.v0.models.commons.model import BulkResponse
from argilla_server.apis.v0.models.text2text import (
    Text2TextBulkRequest,
    Text2TextRecordInputs,
    Text2TextSearchResults,
)
from argilla_server.commons.models import TaskType
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import User

if TYPE_CHECKING:
    from httpx import AsyncClient


@pytest.mark.asyncio
async def test_search_records(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_search_records"
    await delete_dataset(async_client, dataset, workspace_name=argilla_user.username)

    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "task": TaskType.text2text.value, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    records = [
        Text2TextRecordInputs.parse_obj(data)
        for data in [
            {
                "id": 0,
                "text": "This is a text data",
                "metadata": {"field_one": "value one"},
                "prediction": {"agent": "test", "sentences": [{"text": "This is a test data", "score": 0.6}]},
            },
            {"id": 1, "text": "Ånother data"},
        ]
    ]
    response = await async_client.post(
        f"/api/datasets/{dataset}/Text2Text:bulk",
        json=Text2TextBulkRequest(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.dataset == dataset
    assert bulk_response.failed == 0
    assert bulk_response.processed == 2

    response = await async_client.post(
        f"/api/datasets/{dataset}/Text2Text:search", json={}, params=workspace_query_params
    )
    assert response.status_code == 200, response.json()

    results = Text2TextSearchResults.parse_obj(response.json())
    assert results.total == 2
    assert results.records[0].predicted is None

    assert results.aggregations.dict(exclude={"score"}) == {
        "annotated_as": {},
        "annotated_by": {},
        "annotated_text": {},
        "metadata": {"field_one": {"value one": 1}},
        "predicted": {},
        "predicted_as": {},
        "predicted_by": {"test": 1},
        "predicted_text": {},
        "status": {"Default": 2},
        "words": {"data": 2, "text": 1, "ånother": 1},
    }


async def search_data(
    *,
    async_client: "AsyncClient",
    base_url: str,
    workspace_name: str,
    expected_total: int,
    body: Optional[dict] = None,
    vector_name: Optional[str] = None,
):
    response = await async_client.post(url=f"{base_url}:search", json=body or {}, params={"workspace": workspace_name})
    assert response.status_code == 200, response.json()

    results = Text2TextSearchResults.parse_obj(response.json())
    assert results.total == expected_total
    for record in results.records:
        if vector_name:
            assert record.vectors is not None
            assert vector_name in record.vectors


@pytest.mark.asyncio
async def test_search_with_vectors(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_search_with_vectors"

    await delete_dataset(async_client, dataset, workspace_name=argilla_user.username)
    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "task": TaskType.text2text.value, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    records_for_text2text_with_vectors = [
        Text2TextRecordInputs.parse_obj(data)
        for data in [
            {
                "id": 0,
                "text": "This is a text data",
                "metadata": {"field_one": "value one"},
                "prediction": {"agent": "test", "sentences": [{"text": "This is a test data", "score": 0.6}]},
                "vectors": {"my_bert": {"value": [1, 2, 3, 4]}},
            },
            {
                "id": 1,
                "text": "Ånother data",
                "vectors": {"my_bert": {"value": [4, 5, 6, 7]}},
            },
            {
                "id": 3,
                "text": "This is another text data",
                "prediction": {"agent": "test", "sentences": [{"text": "This is another test data", "score": 0.6}]},
            },
        ]
    ]

    base_url = await prepare_data(
        async_client=async_client,
        dataset=dataset,
        workspace_name=argilla_user.username,
        records=records_for_text2text_with_vectors,
    )

    await search_data(
        async_client=async_client,
        workspace_name=argilla_user.username,
        base_url=base_url,
        expected_total=2,
        vector_name="my_bert",
        body={
            "query": {
                "vector": {
                    "name": "my_bert",
                    "value": [1.2, 2.3, 4.1, 6.1],
                }
            }
        },
    )


@pytest.mark.asyncio
async def test_api_with_new_predictions_data_model(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_api_with_new_predictions_data_model"
    await delete_dataset(async_client, dataset, workspace_name=argilla_user.username)
    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "task": TaskType.text2text.value, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    records = [
        Text2TextRecordInputs.parse_obj(
            {
                "text": "This is a text data",
                "predictions": {"test": {"sentences": [{"text": "This is a test data", "score": 0.6}]}},
            }
        ),
        Text2TextRecordInputs.parse_obj(
            {
                "text": "Another data",
                "annotations": {
                    "annotator-1": {"sentences": [{"text": "THis is a test data"}]},
                    "annotator-2": {"sentences": [{"text": "This IS the test datay"}]},
                },
            }
        ),
    ]

    await prepare_data(
        async_client=async_client, dataset=dataset, workspace_name=argilla_user.username, records=records
    )

    response = await async_client.post(
        f"/api/datasets/{dataset}/Text2Text:search",
        json={"query": {"query_text": "predictions.test.sentences.text.exact:data"}},
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    results = Text2TextSearchResults.parse_obj(response.json())
    assert results.total == 1, results

    response = await async_client.post(
        f"/api/datasets/{dataset}/Text2Text:search",
        json={"query": {"query_text": "_exists_:annotations.annotator-1"}},
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    results = Text2TextSearchResults.parse_obj(response.json())
    assert results.total == 1, results


async def prepare_data(
    *, async_client: "AsyncClient", dataset: str, workspace_name: str, records: List[Text2TextRecordInputs]
):
    base_api_url = f"/api/datasets/{dataset}/Text2Text"
    response = await async_client.post(
        f"{base_api_url}:bulk",
        json=Text2TextBulkRequest(
            records=records,
        ).dict(by_alias=True),
        params={"workspace": workspace_name},
    )
    assert response.status_code == 200, response.json()

    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.dataset == dataset
    assert bulk_response.failed == 0
    assert bulk_response.processed == len(records)

    return base_api_url


async def delete_dataset(async_client: "AsyncClient", dataset: str, workspace_name: str):
    response = await async_client.delete(f"/api/datasets/{dataset}", params={"workspace": workspace_name})
    assert response.status_code == 200



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import List, Optional
from uuid import UUID

from argilla_server.enums import DatasetStatus
from argilla_server.pydantic_v1 import BaseModel, Field, constr
from argilla_server.schemas.base import UpdateSchema

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

DATASET_NAME_REGEX = r"^(?!-|_)[a-zA-Z0-9-_ ]+$"
DATASET_NAME_MIN_LENGTH = 1
DATASET_NAME_MAX_LENGTH = 200
DATASET_GUIDELINES_MIN_LENGTH = 1
DATASET_GUIDELINES_MAX_LENGTH = 10000


DatasetName = Annotated[
    constr(regex=DATASET_NAME_REGEX, min_length=DATASET_NAME_MIN_LENGTH, max_length=DATASET_NAME_MAX_LENGTH),
    Field(..., description="Dataset name"),
]


DatasetGuidelines = Annotated[
    constr(min_length=DATASET_GUIDELINES_MIN_LENGTH, max_length=DATASET_GUIDELINES_MAX_LENGTH),
    Field(..., description="Dataset guidelines"),
]


class RecordMetrics(BaseModel):
    count: int


class ResponseMetrics(BaseModel):
    count: int
    submitted: int
    discarded: int
    draft: int


class DatasetMetrics(BaseModel):
    records: RecordMetrics
    responses: ResponseMetrics


class Dataset(BaseModel):
    id: UUID
    name: str
    guidelines: Optional[str]
    allow_extra_metadata: bool
    status: DatasetStatus
    workspace_id: UUID
    last_activity_at: datetime
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True


class Datasets(BaseModel):
    items: List[Dataset]


class DatasetCreate(BaseModel):
    name: DatasetName
    guidelines: Optional[DatasetGuidelines]
    allow_extra_metadata: bool = True
    workspace_id: UUID


class DatasetUpdate(UpdateSchema):
    name: Optional[DatasetName]
    guidelines: Optional[DatasetGuidelines]
    allow_extra_metadata: Optional[bool]

    __non_nullable_fields__ = {"name", "allow_extra_metadata"}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import contextlib
from uuid import uuid4

import pytest
from argilla_server.commons.models import TaskType
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import UserRole

from tests.factories import AnnotatorFactory, DatasetFactory, UserFactory, WorkspaceFactory


@contextlib.asynccontextmanager
async def create_old_argilla_dataset(async_client: "AsyncClient", name: str, workspace_name: str, task: TaskType):
    try:
        response = await async_client.post(
            "/api/datasets", json={"name": name, "workspace": workspace_name, "task": task.value}
        )
        assert response.status_code == 200

        yield response.json()
    finally:
        response = await async_client.delete(f"/api/datasets/{name}?workspace={workspace_name}")
        assert response.status_code == 200


@pytest.mark.asyncio
class TestSuiteWorkspaces:
    async def test_get_workspace(self, async_client: "AsyncClient", owner_auth_header: dict):
        workspace = await WorkspaceFactory.create(name="workspace")

        response = await async_client.get(f"/api/v1/workspaces/{workspace.id}", headers=owner_auth_header)

        assert response.status_code == 200
        assert response.json() == {
            "id": str(workspace.id),
            "name": "workspace",
            "inserted_at": workspace.inserted_at.isoformat(),
            "updated_at": workspace.updated_at.isoformat(),
        }

    async def test_get_workspace_without_authentication(self, async_client: "AsyncClient"):
        workspace = await WorkspaceFactory.create()

        response = await async_client.get(f"/api/v1/workspaces/{workspace.id}")

        assert response.status_code == 401

    async def test_get_workspace_as_annotator(self, async_client: "AsyncClient"):
        workspace = await WorkspaceFactory.create(name="workspace")
        annotator = await AnnotatorFactory.create(workspaces=[workspace])

        response = await async_client.get(
            f"/api/v1/workspaces/{workspace.id}", headers={API_KEY_HEADER_NAME: annotator.api_key}
        )

        assert response.status_code == 200
        assert response.json()["name"] == "workspace"

    async def test_get_workspace_as_annotator_from_different_workspace(self, async_client: "AsyncClient"):
        workspace = await WorkspaceFactory.create()
        another_workspace = await WorkspaceFactory.create()
        annotator = await AnnotatorFactory.create(workspaces=[another_workspace])

        response = await async_client.get(
            f"/api/v1/workspaces/{workspace.id}", headers={API_KEY_HEADER_NAME: annotator.api_key}
        )

        assert response.status_code == 403

    async def test_get_workspace_with_nonexistent_workspace_id(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        await WorkspaceFactory.create()

        response = await async_client.get(f"/api/v1/workspaces/{uuid4()}", headers=owner_auth_header)

        assert response.status_code == 404

    async def test_delete_workspace(self, async_client: "AsyncClient", owner_auth_header: dict):
        workspace = await WorkspaceFactory.create(name="workspace_delete")
        other_workspace = await WorkspaceFactory.create()

        await DatasetFactory.create_batch(3, workspace=other_workspace)

        response = await async_client.delete(f"/api/v1/workspaces/{workspace.id}", headers=owner_auth_header)

        assert response.status_code == 200

    async def test_delete_workspace_with_feedback_datasets(self, async_client: "AsyncClient", owner_auth_header: dict):
        workspace = await WorkspaceFactory.create(name="workspace_delete")

        await DatasetFactory.create_batch(3, workspace=workspace)

        response = await async_client.delete(f"/api/v1/workspaces/{workspace.id}", headers=owner_auth_header)

        assert response.status_code == 409
        assert response.json() == {
            "detail": f"Cannot delete the workspace {workspace.id}. This workspace has some feedback datasets linked"
        }

    @pytest.mark.parametrize("task", [TaskType.text_classification, TaskType.token_classification, TaskType.text2text])
    async def test_delete_workspace_with_old_datasets(
        self, async_client: "AsyncClient", owner_auth_header: dict, task: TaskType
    ):
        workspace = await WorkspaceFactory.create(name="workspace_delete")

        async_client.headers.update(owner_auth_header)
        async with create_old_argilla_dataset(async_client, name="dataset", workspace_name=workspace.name, task=task):
            response = await async_client.delete(f"/api/v1/workspaces/{workspace.id}")

            assert response.status_code == 409
            assert response.json() == {
                "detail": f"Cannot delete the workspace {workspace.id}. This workspace has some datasets linked"
            }

    async def test_delete_missing_workspace(self, async_client: "AsyncClient", owner_auth_header: dict):
        async_client.headers.update(owner_auth_header)
        response = await async_client.delete(f"/api/v1/workspaces/{uuid4()}")

        assert response.status_code == 404

    @pytest.mark.parametrize("role", [UserRole.annotator, UserRole.admin])
    async def test_delete_workspace_without_permissions(self, async_client: "AsyncClient", role: UserRole):
        workspace = await WorkspaceFactory.create(name="workspace_delete")

        user = await UserFactory.create(role=role, workspaces=[workspace])
        async_client.headers.update({API_KEY_HEADER_NAME: user.api_key})
        response = await async_client.delete(f"/api/v1/workspaces/{workspace.id}")

        assert response.status_code == 403

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin, UserRole.annotator])
    async def test_list_workspaces_me(self, async_client: "AsyncClient", role: UserRole) -> None:
        workspaces = await WorkspaceFactory.create_batch(size=5)
        user = await UserFactory.create(role=role, workspaces=workspaces if role != UserRole.owner else [])

        response = await async_client.get("/api/v1/me/workspaces", headers={API_KEY_HEADER_NAME: user.api_key})

        assert response.status_code == 200
        assert len(response.json()["items"]) == len(workspaces)
        for workspace in workspaces:
            assert {
                "id": str(workspace.id),
                "name": workspace.name,
                "inserted_at": workspace.inserted_at.isoformat(),
                "updated_at": workspace.updated_at.isoformat(),
            } in response.json()["items"]

    async def test_list_workspaces_me_without_authentication(self, async_client: "AsyncClient") -> None:
        response = await async_client.get("/api/v1/me/workspaces")

        assert response.status_code == 401

    @pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin, UserRole.annotator])
    async def test_list_workspaces_me_no_workspaces(self, async_client: "AsyncClient", role: UserRole) -> None:
        user = await UserFactory.create(role=role)

        response = await async_client.get("/api/v1/me/workspaces", headers={API_KEY_HEADER_NAME: user.api_key})

        assert response.status_code == 200
        assert len(response.json()["items"]) == 0



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union

import pytest
from argilla_server.schemas.v0.users import User, UserCreate
from argilla_server.schemas.v0.workspaces import WorkspaceCreate

from tests.factories import UserFactory, WorkspaceFactory
from tests.pydantic_v1 import ValidationError


@pytest.mark.parametrize("invalid_name", ["work space", "work/space", "work.space", "_", "-"])
def test_workspace_create_invalid_name(invalid_name: str):
    with pytest.raises(ValidationError):
        WorkspaceCreate(name=invalid_name)


@pytest.mark.parametrize(
    "username",
    [
        "user-name",
        "user_name",
        "user123",
        "user-123",
        "user_123",
        "UserName",
        "userName",
        "User_name",
        "valid_user_name",
        "user-123_abc",
        "user_123-abc",
        "0033_user",
        "12-user",
    ],
)
def test_user_create(username: str):
    assert UserCreate(first_name="first-name", username=username, password="12345678")


@pytest.mark.parametrize(
    "invalid_username",
    [
        "user name",
        "user/name",
        "user.name",
        "_",
        "-",
        "-1234",
        "_1234",
        "_mark",
        "-mark",
    ],
)
def test_user_create_invalid_username(invalid_username: str):
    with pytest.raises(ValidationError):
        UserCreate(first_name="first-name", username=invalid_username, password="12345678")


@pytest.mark.asyncio
async def test_user_first_name():
    user = await UserFactory.create(first_name="first-name", workspaces=[])

    assert User.from_orm(user).first_name == "first-name"


@pytest.mark.asyncio
async def test_user_last_name():
    user = await UserFactory.create(last_name="last-name", workspaces=[])

    assert User.from_orm(user).last_name == "last-name"


@pytest.mark.asyncio
async def test_user_full_name():
    user = await UserFactory.create(first_name="first-name", last_name="last-name", workspaces=[])

    assert User.from_orm(user).full_name == "first-name last-name"


@pytest.mark.parametrize("last_name", [None, ""])
@pytest.mark.asyncio
async def test_user_full_name_without_last_name(last_name: Union[str, None]):
    user = await UserFactory.create(first_name="first-name", last_name=last_name, workspaces=[])

    assert User.from_orm(user).full_name == "first-name"


@pytest.mark.asyncio
async def test_user_workspaces():
    workspace_a = await WorkspaceFactory.create(name="workspace-a")
    workspace_b = await WorkspaceFactory.create(name="workspace-b")
    user = await UserFactory.create(workspaces=[workspace_a, workspace_b])

    assert User.from_orm(user).workspaces == ["workspace-a", "workspace-b"]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, Optional, Type

from fastapi import Depends

from argilla_server.daos.backend.search.model import SortConfig
from argilla_server.daos.models.records import DaoRecordsSearch
from argilla_server.daos.records import DatasetRecordsDAO
from argilla_server.services.datasets import ServiceDataset
from argilla_server.services.metrics.models import ServiceMetric, ServicePythonMetric
from argilla_server.services.search.model import ServiceBaseRecordsQuery, ServiceRecordsQuery
from argilla_server.services.tasks.commons import ServiceRecord


class MetricsService:
    """The dataset metrics service singleton"""

    _INSTANCE = None

    @classmethod
    def get_instance(
        cls,
        dao: DatasetRecordsDAO = Depends(DatasetRecordsDAO.get_instance),
    ) -> "MetricsService":
        """
        Creates the service instance.

        Parameters
        ----------
        dao:
            The dataset records dao

        Returns
        -------
            The metrics service instance

        """
        if not cls._INSTANCE:
            cls._INSTANCE = cls(dao)
        return cls._INSTANCE

    def __init__(self, dao: DatasetRecordsDAO):
        """
        Creates a service instance

        Parameters
        ----------
        dao:
            The dataset records dao
        """
        self.__dao__ = dao

    def summarize_metric(
        self,
        dataset: ServiceDataset,
        metric: ServiceMetric,
        record_class: Optional[Type[ServiceRecord]] = None,
        query: Optional[ServiceRecordsQuery] = None,
        **metric_params,
    ) -> Dict[str, Any]:
        """
        Applies a metric summarization.

        Parameters
        ----------
        dataset:
            The records dataset
        metric:
            The selected metric
        record_class:
            The record class type for python metrics computation
        query:
            An optional query passed for records filtering
        metric_params:
            Related metrics parameters

        Returns
        -------
            The metric summarization info
        """

        if isinstance(metric, ServicePythonMetric):
            query = query or ServiceBaseRecordsQuery()
            query = metric.prepare_query(query)
            records = self.__dao__.scan_dataset(
                dataset,
                search=DaoRecordsSearch(query=query, sort=SortConfig(shuffle=metric.shuffle_records)),
                limit=metric.records_to_fetch,
                exclude_fields={
                    "vectors",
                    "metrics",
                    "metadata",
                },  # Load records more efficiently
            )
            return metric.apply(map(record_class.parse_obj, records))

        return self.__dao__.compute_metric(
            metric_id=metric.id,
            metric_params=metric_params,
            dataset=dataset,
            query=query,
        )

    def annotated_records(self, dataset: ServiceDataset) -> int:
        """Return the number of annotated records for a dataset"""
        results = self.__dao__.search_records(
            dataset,
            size=0,
            search=DaoRecordsSearch(query=ServiceBaseRecordsQuery(has_annotation=True)),
        )
        return results.total

    def total_records(self, dataset: ServiceDataset) -> int:
        """Return the total number of records for a given dataset"""
        return self.__dao__.search_records(dataset, size=0).total



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Annotated, List, Optional
from uuid import UUID

from typing_extensions import Annotated

from argilla_server.pydantic_v1 import BaseModel, Field, PositiveInt, constr
from argilla_server.schemas.base import UpdateSchema

VECTOR_SETTINGS_CREATE_NAME_REGEX = r"^(?=.*[a-z0-9])[a-z0-9_-]+$"
VECTOR_SETTINGS_CREATE_NAME_MIN_LENGTH = 1
VECTOR_SETTINGS_CREATE_NAME_MAX_LENGTH = 200

VECTOR_SETTINGS_CREATE_TITLE_MIN_LENGTH = 1
VECTOR_SETTINGS_CREATE_TITLE_MAX_LENGTH = 500


VectorSettingsTitle = Annotated[
    constr(
        min_length=VECTOR_SETTINGS_CREATE_TITLE_MIN_LENGTH,
        max_length=VECTOR_SETTINGS_CREATE_TITLE_MAX_LENGTH,
    ),
    Field(..., description="The title of the vector settings"),
]


class VectorSettings(BaseModel):
    id: UUID
    name: str
    title: str
    dimensions: int
    dataset_id: UUID
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True

    def check_vector(self, value: List[float]) -> None:
        num_elements = len(value)

        if num_elements != self.dimensions:
            raise ValueError(f"vector must have {self.dimensions} elements, got {num_elements} elements")


class VectorsSettings(BaseModel):
    items: List[VectorSettings]


class VectorSettingsCreate(BaseModel):
    name: str = Field(
        ...,
        regex=VECTOR_SETTINGS_CREATE_NAME_REGEX,
        min_length=VECTOR_SETTINGS_CREATE_NAME_MIN_LENGTH,
        max_length=VECTOR_SETTINGS_CREATE_NAME_MAX_LENGTH,
        description="The title of the vector settings",
    )
    title: VectorSettingsTitle
    dimensions: PositiveInt


class VectorSettingsUpdate(UpdateSchema):
    title: Optional[VectorSettingsTitle]

    __non_nullable_fields__ = {"title"}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import cast

import pytest
from argilla_server._app import create_server_app
from argilla_server.settings import Settings, settings
from starlette.routing import Mount
from starlette.testclient import TestClient


@pytest.fixture
def test_settings():
    yield settings

    settings.base_url = "/"


class TestApp:
    def test_create_app_with_base_url(self, test_settings: Settings):
        base_url = "/base/url"
        settings.base_url = base_url

        app = create_server_app()
        client = TestClient(app)

        response = client.get("/api/docs")
        assert response.status_code == 404

        response = client.get(f"{base_url}/api/docs")
        assert response.status_code == 200

        response = client.get("/api/_info")
        assert response.status_code == 404

        response = client.get(f"{base_url}/api/_info")
        assert response.status_code == 200

        assert len(app.routes) == 1
        assert cast(Mount, app.routes[0]).path == base_url



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""update_enum_columns

Revision ID: 8c574ada5e5f
Revises: 3ff6484f8b37
Create Date: 2023-06-23 15:55:19.928164

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8c574ada5e5f"
down_revision = "3ff6484f8b37"
branch_labels = None
depends_on = None

# Aligned with the values of `ResponseStatus` in `src/argilla/server/models/dataset.py`
response_status_enum = sa.Enum("draft", "submitted", "discarded", name="response_status_enum")

# Aligned with the values of `DatasetStatus` in `src/argilla/server/models/dataset.py`
dataset_status_enum = sa.Enum("draft", "ready", name="dataset_status_enum")

# Aligned with the values of `UserRole` in `src/argilla/server/models/dataset.py`
user_role_enum = sa.Enum("owner", "admin", "annotator", name="user_role_enum")


def upgrade() -> None:
    bind = op.get_bind()

    if bind.dialect.name == "postgresql":
        response_status_enum.create(bind)
        op.execute(
            "ALTER TABLE responses ALTER COLUMN status TYPE response_status_enum USING status::response_status_enum"
        )

        dataset_status_enum.create(bind)
        op.execute(
            "ALTER TABLE datasets ALTER COLUMN status TYPE dataset_status_enum USING status::dataset_status_enum"
        )

        user_role_enum.create(bind)
        op.execute("ALTER TABLE users ALTER COLUMN role TYPE user_role_enum USING role::user_role_enum")


def downgrade() -> None:
    bind = op.get_bind()

    if bind.dialect.name == "postgresql":
        with op.batch_alter_table("users") as batch_op:
            batch_op.alter_column("role", existing_type=user_role_enum, type_=sa.String(), nullable=False)
        user_role_enum.drop(bind)

        with op.batch_alter_table("datasets") as batch_op:
            batch_op.alter_column("status", existing_type=dataset_status_enum, type_=sa.String(), nullable=False)
        dataset_status_enum.drop(bind)

        with op.batch_alter_table("responses") as batch_op:
            batch_op.alter_column("status", existing_type=response_status_enum, type_=sa.String(), nullable=False)
        response_status_enum.drop(bind)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, List, Set

import pytest
from argilla_server.utils import parse_query_param
from fastapi import HTTPException

from tests.pydantic_v1 import BaseModel, Field


@pytest.mark.parametrize(
    "config, params, expected",
    [
        ({}, ["key1", "key2", "key3"], {"keys": ["key1", "key2", "key3"]}),
        ({}, ["key1,key2,key3"], {"keys": ["key1", "key2", "key3"]}),
        ({}, ["key1:value1,value2,value3"], {"key1": ["value1", "value2", "value3"]}),
        (
            {},
            ["key1:value1,value2,value3", "key2:value4,value5,value6"],
            {"key1": ["value1", "value2", "value3"], "key2": ["value4", "value5", "value6"]},
        ),
        ({}, ["key1", "key2", "key3:value1,value2"], {"keys": ["key1", "key2"], "key3": ["value1", "value2"]}),
        ({}, ["key1:value1", "key1:value2"], {"key1": ["value1", "value2"]}),
        ({"max_keys": 1}, ["key1"], {"key": "key1"}),
        ({"max_keys": 1}, ["key1:value1,value2,value3"], {"key": "key1", "value": ["value1", "value2", "value3"]}),
        ({"max_keys": 1, "max_values_per_key": 1}, ["key1:value1"], {"key": "key1", "value": "value1"}),
        ({"max_values_per_key": 1}, ["key1:value1", "key2:value2"], {"key1": "value1", "key2": "value2"}),
        ({"max_values_per_key": 1, "group_keys_without_values": False}, ["key1", "key2"], {"key1": None, "key2": None}),
        ({"group_keys_without_values": False}, ["key1,key2,key3"], {"key1": None, "key2": None, "key3": None}),
        ({"group_keys_without_values": True}, ["key1,key2,key3"], {"keys": ["key1", "key2", "key3"]}),
    ],
)
def test_parse_query_param(config: Dict[str, Any], params: List[str], expected: Dict[str, Set[str]]) -> None:
    parse_function = parse_query_param(name="unit-test", **config)
    result = parse_function(param_values=params)
    assert result == expected


def test_parse_query_param_with_base_model() -> None:
    class Params(BaseModel):
        keys: List[str] = Field(..., alias="keys")
        my_other_key_with_values: List[str] = Field(..., alias="key5")

    parse_function = parse_query_param(name="unit-test", model=Params)
    result = parse_function(param_values=["key1", "key2", "key3,key4", "key5:value1,value2,value3"])
    assert sorted(result.keys) == ["key1", "key2", "key3", "key4"]
    assert sorted(result.my_other_key_with_values) == ["value1", "value2", "value3"]


@pytest.mark.parametrize(
    "config, params, expected_msg",
    [
        (
            {},
            ["key1,key2,key3:value1,value2,value3"],
            "'unit-test' query parameter must be of the form 'key1,key2,key3' or 'key:value1,value2,value3'",
        ),
        ({"max_keys": 1}, ["key1,key2"], "'unit-test' query parameter must contain at most 1 comma-separated keys"),
        (
            {"max_values_per_key": 2},
            ["key1:value1,value2,value3"],
            "'unit-test' query parameter must contain at most 2 values per comma-separated key. 'key1' has 3 values.",
        ),
    ],
)
def test_parse_query_param_raises_http_exception(config: Dict[str, Any], params: List[str], expected_msg: str) -> None:
    parse_function = parse_query_param(name="unit-test", **config)

    with pytest.raises(HTTPException, match="") as exc_info:
        parse_function(param_values=params)

    assert exc_info.value.status_code == 422
    assert exc_info.value.detail == expected_msg



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from httpx import AsyncClient


@pytest.mark.asyncio
async def test_docs_redirect(async_client: "AsyncClient"):
    response = await async_client.get("/docs", follow_redirects=False)
    assert response.status_code == 307
    assert response.next_request.url.path == "/api/docs"

    response = await async_client.get("/api", follow_redirects=False)
    assert response.status_code == 307
    assert response.next_request.url.path == "/api/docs"



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import ClassVar, List, Optional

from fastapi import Depends, FastAPI
from fastapi.security import SecurityScopes
from sqlalchemy.ext.asyncio import AsyncSession
from starlette.authentication import AuthenticationBackend
from starlette.requests import Request

from argilla_server.contexts import accounts
from argilla_server.database import get_async_db
from argilla_server.errors import UnauthorizedError
from argilla_server.models import User
from argilla_server.security.authentication.db import APIKeyAuthenticationBackend, BearerTokenAuthenticationBackend
from argilla_server.security.authentication.userinfo import UserInfo


class AuthenticationProvider:
    """Authentication provider for the API requests."""

    backends: ClassVar[List[AuthenticationBackend]] = [
        APIKeyAuthenticationBackend(),
        BearerTokenAuthenticationBackend(),
    ]

    @classmethod
    def new_instance(cls):
        return AuthenticationProvider()

    def configure_app(self, app: FastAPI) -> None:
        pass

    async def get_current_user(
        self,
        security_scopes: SecurityScopes,  # noqa
        request: Request,
        db: AsyncSession = Depends(get_async_db),
        _api_key: Optional[str] = Depends(APIKeyAuthenticationBackend.scheme),
        _bearer: Optional[str] = Depends(BearerTokenAuthenticationBackend.scheme),
    ) -> User:
        """Get the current user from the request."""

        userinfo = await self._authenticate_request_user(db, request)
        if not userinfo:
            raise UnauthorizedError()

        user = await accounts.get_user_by_username(db, userinfo.username)
        if not user:
            raise UnauthorizedError()

        return user

    async def _authenticate_request_user(self, db: AsyncSession, request: Request) -> Optional[UserInfo]:
        # This db will be used by the backends. Ideally this should be done as a global dependency
        # but is not working as expected. Sometimes the db is not available in the request state at the
        # middlewares level.
        request.state.db = db

        for backend in self.backends:
            if results := await backend.authenticate(request):
                _, userinfo = results
                return userinfo

        return None



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from typing import TYPE_CHECKING
from unittest import mock

from argilla_server.models import User, UserRole, Workspace, WorkspaceUser
from click.testing import CliRunner
from typer import Typer

if TYPE_CHECKING:
    from sqlalchemy.orm import Session


def test_migrate(monkeypatch, sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    mock_users_file = os.path.join(os.path.dirname(__file__), "test_user_files", "users.yml")

    with mock.patch.dict(os.environ, {"ARGILLA_LOCAL_AUTH_USERS_DB_FILE": mock_users_file}):
        result = cli_runner.invoke(cli, "database users migrate")

        assert result.exit_code == 0
        assert sync_db.query(User).count() == 5
        assert sync_db.query(Workspace).count() == 9
        assert sync_db.query(WorkspaceUser).count() == 11

        user = sync_db.query(User).filter_by(username="john").first()
        assert user.first_name == "John Doe"
        assert user.username == "john"
        assert user.role == UserRole.owner
        assert user.api_key == "a14427ea-9197-11ec-b909-0242ac120002"
        assert user.password_hash == "$2y$05$xtl7iy3bpqchUwiQMjEHe.tY7OaIjDrg43W3TB4EHQ7izvdjvGtPS"
        assert [ws.name for ws in user.workspaces] == ["john"]

        user = sync_db.query(User).filter_by(username="tanya").first()
        assert user.first_name == "Tanya Franklin"
        assert user.username == "tanya"
        assert user.role == UserRole.annotator
        assert user.api_key == "78a10b53-8db7-4ab5-9e9e-fbd4b7e76551"
        assert user.password_hash == "$2y$05$aqNyXcXRXddNj5toZwT0HugHqKZypvqlBAkZviAGGbsAC8oTj/P5K"
        assert [ws.name for ws in user.workspaces] == ["tanya", "argilla", "team"]

        user = sync_db.query(User).filter_by(username="daisy").first()
        assert user.first_name == "Daisy Gonzalez"
        assert user.username == "daisy"
        assert user.role == UserRole.annotator
        assert user.api_key == "a8168929-8668-494c-b7a5-98cd35740d9b"
        assert user.password_hash == "$2y$05$l83IhUs4ZDaxsgZ/P12FO.RFTi2wKQ2AxMK2vYtLx//yKramuCcZG"
        assert set([ws.name for ws in user.workspaces]) == {"daisy", "argilla", "team", "latam"}

        user = sync_db.query(User).filter_by(username="macleod").first()
        assert user.first_name == ""
        assert user.username == "macleod"
        assert user.role == UserRole.annotator
        assert user.api_key == "7c3b4d6e-1898-4c42-84c8-e1758cea1ce0"
        assert user.password_hash == "$2y$05$Fb3iv7AGv8k.o5cl9qdCtuwkrLcDcSYKWyJk1QNl6RXKUecvP.Ium"
        assert [ws.name for ws in user.workspaces] == ["macleod", "highlands"]

        user = sync_db.query(User).filter_by(username="sanchez").first()
        assert user.first_name == "Juan Sánchez Villalobos Ramírez"
        assert user.username == "sanchez"
        assert user.role == UserRole.annotator
        assert user.api_key == "ac7b6b86-7d63-45ce-a76a-08f64e0d5fd6"
        assert user.password_hash == "$2y$05$wMvfoz2TwrRFRZhNELHjbOcqEucVYImNORuRvh7Vp26.dIqvo9tY2"
        assert [ws.name for ws in user.workspaces] == ["sanchez"]


def test_migrate_with_one_user_file(monkeypatch, sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    mock_users_file = os.path.join(os.path.dirname(__file__), "test_user_files", "users_one.yml")

    with mock.patch.dict(os.environ, {"ARGILLA_LOCAL_AUTH_USERS_DB_FILE": mock_users_file}):
        result = cli_runner.invoke(cli, "database users migrate")

        assert result.exit_code == 0
        assert sync_db.query(User).count() == 1
        assert sync_db.query(Workspace).count() == 3
        assert sync_db.query(WorkspaceUser).count() == 3

        user = sync_db.query(User).filter_by(username="john").first()
        assert user.first_name == "John Doe"
        assert user.username == "john"
        assert user.role == UserRole.annotator
        assert user.api_key == "a14427ea-9197-11ec-b909-0242ac120002"
        assert user.password_hash == "$2y$05$xtl7iy3bpqchUwiQMjEHe.tY7OaIjDrg43W3TB4EHQ7izvdjvGtPS"
        assert [ws.name for ws in user.workspaces] == ["john", "argilla", "team"]


def test_migrate_with_invalid_user(monkeypatch, sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    mock_users_file = os.path.join(os.path.dirname(__file__), "test_user_files", "users_invalid_user.yml")

    with mock.patch.dict(os.environ, {"ARGILLA_LOCAL_AUTH_USERS_DB_FILE": mock_users_file}):
        result = cli_runner.invoke(cli, "database users migrate")

        assert result.exit_code == 1
        assert sync_db.query(User).count() == 0
        assert sync_db.query(Workspace).count() == 0
        assert sync_db.query(WorkspaceUser).count() == 0


def test_migrate_with_invalid_workspace(monkeypatch, sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    mock_users_file = os.path.join(os.path.dirname(__file__), "test_user_files", "users_invalid_workspace.yml")

    with mock.patch.dict(os.environ, {"ARGILLA_LOCAL_AUTH_USERS_DB_FILE": mock_users_file}):
        result = cli_runner.invoke(cli, "database users migrate")

        assert result.exit_code == 1
        assert sync_db.query(User).count() == 0
        assert sync_db.query(Workspace).count() == 0
        assert sync_db.query(WorkspaceUser).count() == 0


def test_migrate_with_nonexistent_file(monkeypatch, sync_db: "Session", cli_runner: CliRunner, cli: Typer):
    with mock.patch.dict(os.environ, {"ARGILLA_LOCAL_AUTH_USERS_DB_FILE": "nonexistent.yml"}):
        result = cli_runner.invoke(cli, "database users migrate")

        assert result.exit_code == 1
        assert sync_db.query(User).count() == 0
        assert sync_db.query(Workspace).count() == 0
        assert sync_db.query(WorkspaceUser).count() == 0



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server.contexts import datasets
from argilla_server.database import get_async_db
from argilla_server.models import User, VectorSettings
from argilla_server.policies import VectorSettingsPolicyV1, authorize
from argilla_server.schemas.v1.vector_settings import VectorSettings as VectorSettingsSchema
from argilla_server.schemas.v1.vector_settings import VectorSettingsUpdate
from argilla_server.security import auth

router = APIRouter(tags=["vectors-settings"])


async def _get_vector_settings(db: AsyncSession, vector_settings_id: UUID) -> VectorSettings:
    vector_settings = await datasets.get_vector_settings_by_id(db, vector_settings_id)
    if not vector_settings:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Vector settings with `{vector_settings_id}` not found",
        )
    return vector_settings


@router.patch("/vectors-settings/{vector_settings_id}", response_model=VectorSettingsSchema)
async def update_vector_settings(
    *,
    db: AsyncSession = Depends(get_async_db),
    vector_settings_id: UUID,
    vector_settings_update: VectorSettingsUpdate,
    current_user: User = Security(auth.get_current_user),
):
    vector_settings = await _get_vector_settings(db, vector_settings_id)

    await authorize(current_user, VectorSettingsPolicyV1.update(vector_settings))

    return await datasets.update_vector_settings(db, vector_settings, vector_settings_update)


@router.delete("/vectors-settings/{vector_settings_id}", response_model=VectorSettingsSchema)
async def delete_vector_settings(
    *,
    db: AsyncSession = Depends(get_async_db),
    vector_settings_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    vector_settings = await _get_vector_settings(db, vector_settings_id)

    await authorize(current_user, VectorSettingsPolicyV1.delete(vector_settings))

    return await datasets.delete_vector_settings(db, vector_settings)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, Generic, List, Literal, Optional, TypeVar, Union

from argilla_server.enums import MetadataPropertyType
from argilla_server.pydantic_v1 import BaseModel, Field
from argilla_server.pydantic_v1.generics import GenericModel

__all__ = [
    "MetadataPropertySettings",
    "TermsMetadataPropertySettings",
    "IntegerMetadataPropertySettings",
    "FloatMetadataPropertySettings",
]

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated


class BaseMetadataPropertySettings(BaseModel, ABC):
    @abstractmethod
    def check_metadata(self, value: Any) -> None:
        pass


class TermsMetadataPropertySettings(BaseMetadataPropertySettings):
    type: Literal[MetadataPropertyType.terms]
    values: Optional[List[str]] = None

    def check_metadata(self, value: Union[str, List[str]]) -> None:
        if self.values is None:
            return

        values = value
        if isinstance(values, str):
            values = [value]

        for v in values:
            if v not in self.values:
                raise ValueError(f"'{v}' is not an allowed term.")


NT = TypeVar("NT", int, float)


class NumericMetadataPropertySettings(BaseMetadataPropertySettings, GenericModel, Generic[NT]):
    min: Optional[NT] = None
    max: Optional[NT] = None

    def check_metadata(self, value: NT) -> None:
        if self.min is not None and value < self.min:
            raise ValueError(f"'{value}' is less than the minimum value of '{self.min}'.")

        if self.max is not None and value > self.max:
            raise ValueError(f"'{value}' is greater than the maximum value of '{self.max}'.")


class IntegerMetadataPropertySettings(NumericMetadataPropertySettings[int]):
    type: Literal[MetadataPropertyType.integer]

    def check_metadata(self, value: int) -> None:
        if not isinstance(value, int):
            raise ValueError(f"'{value}' is not an integer.")
        return super().check_metadata(value)


class FloatMetadataPropertySettings(NumericMetadataPropertySettings[float]):
    type: Literal[MetadataPropertyType.float]

    def check_metadata(self, value: float) -> None:
        if not isinstance(value, float):
            raise ValueError(f"'{value}' is not a float.")
        return super().check_metadata(value)


MetadataPropertySettings = Annotated[
    Union[
        TermsMetadataPropertySettings,
        IntegerMetadataPropertySettings,
        FloatMetadataPropertySettings,
    ],
    Field(..., discriminator="type"),
]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest
from argilla_server.models import User
from fastapi.testclient import TestClient


@pytest.mark.asyncio
async def test_create_security_token(async_client: "AsyncClient", owner):
    response = await async_client.post("/api/security/token", data={"username": owner.username, "password": "1234"})

    assert response.status_code == 200

    response_body = response.json()
    assert response_body["access_token"]
    assert response_body["token_type"] == "bearer"


@pytest.mark.asyncio
async def test_create_security_token_with_empty_username(async_client: "AsyncClient"):
    response = await async_client.post("/api/security/token", data={"username": "", "password": "1234"})

    assert response.status_code == 422


@pytest.mark.asyncio
async def test_create_security_token_with_empty_password(async_client: "AsyncClient", owner):
    response = await async_client.post("/api/security/token", data={"username": owner.username, "password": ""})

    assert response.status_code == 422


@pytest.mark.asyncio
async def test_create_security_token_with_invalid_username(async_client: "AsyncClient"):
    response = await async_client.post("/api/security/token", data={"username": "invalid", "password": "1234"})

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_create_security_token_with_invalid_password(async_client: "AsyncClient", owner):
    response = await async_client.post("/api/security/token", data={"username": owner.username, "password": "invalid"})

    assert response.status_code == 401



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID, uuid4

import pytest
from argilla_server.apis.v1.handlers.datasets.records import LIST_DATASET_RECORDS_LIMIT_LE
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.enums import RecordInclude, SortOrder
from argilla_server.search_engine import (
    AndFilter,
    Order,
    RangeFilter,
    ResponseFilterScope,
    SearchEngine,
    SearchResponseItem,
    SearchResponses,
    SuggestionFilterScope,
    TermsFilter,
)
from httpx import AsyncClient

from tests.factories import (
    AdminFactory,
    AnnotatorFactory,
    DatasetFactory,
    OwnerFactory,
    RatingQuestionFactory,
    RecordFactory,
    ResponseFactory,
    TextFieldFactory,
    VectorFactory,
    VectorSettingsFactory,
)


@pytest.mark.asyncio
class TestSearchDatasetRecords:
    def url(self, dataset_id: UUID) -> str:
        return f"/api/v1/datasets/{dataset_id}/records/search"

    async def test_as_owner(self, async_client: AsyncClient):
        dataset = await DatasetFactory.create()
        owner = await OwnerFactory.create(workspaces=[dataset.workspace])

        response = await async_client.post(
            self.url(dataset.id),
            headers={API_KEY_HEADER_NAME: owner.api_key},
            json={"query": {"text": {"q": "text"}}},
        )

        assert response.status_code == 200

    async def test_as_admin_from_different_workspace(self, async_client: AsyncClient):
        dataset = await DatasetFactory.create()
        admin = await AdminFactory.create()

        response = await async_client.post(
            self.url(dataset.id),
            headers={API_KEY_HEADER_NAME: admin.api_key},
            json={"query": {"text": {"q": "text"}}},
        )

        assert response.status_code == 403

    async def test_as_annotator(self, async_client: AsyncClient):
        dataset = await DatasetFactory.create()
        annotator = await AnnotatorFactory.create(workspaces=[dataset.workspace])

        response = await async_client.post(
            self.url(dataset.id),
            headers={API_KEY_HEADER_NAME: annotator.api_key},
            json={"query": {"text": {"q": "text"}}},
        )

        assert response.status_code == 403

    async def test_with_include_responses(
        self, async_client: AsyncClient, owner_auth_header: dict, mock_search_engine: SearchEngine
    ):
        dataset = await DatasetFactory.create()
        record_a = await RecordFactory.create(dataset=dataset)
        record_b = await RecordFactory.create(dataset=dataset)

        response_a = await ResponseFactory.create(values={"input_ok": {"value": "yes"}}, record=record_a)
        response_b = await ResponseFactory.create(values={"input_ok": {"value": "no"}}, record=record_a)
        response_c = await ResponseFactory.create(values={"input_ok": {"value": "yes"}}, record=record_b)

        mock_search_engine.search.return_value = SearchResponses(
            items=[
                SearchResponseItem(record_id=record_a.id, score=1.0),
                SearchResponseItem(record_id=record_b.id, score=0.5),
            ],
            total=2,
        )

        response = await async_client.post(
            self.url(dataset.id),
            headers=owner_auth_header,
            params={"include": RecordInclude.responses.value},
            json={"query": {"text": {"q": "text"}}},
        )

        assert response.status_code == 200
        assert response.json() == {
            "items": [
                {
                    "record": {
                        "id": str(record_a.id),
                        "fields": {
                            "sentiment": "neutral",
                            "text": "This is a text",
                        },
                        "metadata": None,
                        "responses": [
                            {
                                "id": str(response_a.id),
                                "status": "submitted",
                                "values": {"input_ok": {"value": "yes"}},
                                "record_id": str(record_a.id),
                                "user_id": str(response_a.user_id),
                                "inserted_at": response_a.inserted_at.isoformat(),
                                "updated_at": response_a.updated_at.isoformat(),
                            },
                            {
                                "id": str(response_b.id),
                                "status": "submitted",
                                "values": {"input_ok": {"value": "no"}},
                                "record_id": str(record_a.id),
                                "user_id": str(response_b.user_id),
                                "inserted_at": response_b.inserted_at.isoformat(),
                                "updated_at": response_b.updated_at.isoformat(),
                            },
                        ],
                        "external_id": record_a.external_id,
                        "dataset_id": str(record_a.dataset_id),
                        "inserted_at": record_a.inserted_at.isoformat(),
                        "updated_at": record_a.updated_at.isoformat(),
                    },
                    "query_score": 1.0,
                },
                {
                    "record": {
                        "id": str(record_b.id),
                        "fields": {
                            "sentiment": "neutral",
                            "text": "This is a text",
                        },
                        "metadata": None,
                        "responses": [
                            {
                                "id": str(response_c.id),
                                "status": "submitted",
                                "values": {"input_ok": {"value": "yes"}},
                                "record_id": str(record_b.id),
                                "user_id": str(response_c.user_id),
                                "inserted_at": response_c.inserted_at.isoformat(),
                                "updated_at": response_c.updated_at.isoformat(),
                            },
                        ],
                        "external_id": record_b.external_id,
                        "dataset_id": str(record_b.dataset_id),
                        "inserted_at": record_b.inserted_at.isoformat(),
                        "updated_at": record_b.updated_at.isoformat(),
                    },
                    "query_score": 0.5,
                },
            ],
            "total": 2,
        }

    async def test_with_invalid_offset(self, async_client: AsyncClient, owner_auth_header: dict):
        response = await async_client.post(
            self.url(uuid4()),
            headers=owner_auth_header,
            params={"offset": -1},
            json={"query": {"text": {"q": "text"}}},
        )

        assert response.status_code == 422

    async def test_with_invalid_lower_limit(self, async_client: AsyncClient, owner_auth_header: dict):
        response = await async_client.post(
            self.url(uuid4()),
            headers=owner_auth_header,
            params={"limit": 0},
            json={"query": {"text": {"q": "text"}}},
        )

        assert response.status_code == 422

    async def test_with_invalid_upper_limit(self, async_client: AsyncClient, owner_auth_header: dict):
        response = await async_client.post(
            self.url(uuid4()),
            headers=owner_auth_header,
            params={"limit": LIST_DATASET_RECORDS_LIMIT_LE + 1},
            json={"query": {"text": {"q": "text"}}},
        )

        assert response.status_code == 422

    async def test_with_non_existent_dataset(self, async_client: AsyncClient, owner_auth_header: dict):
        response = await async_client.post(
            self.url(uuid4()),
            headers=owner_auth_header,
            json={"query": {"text": {"q": "text"}}},
        )

        assert response.status_code == 404

    async def test_with_text_query_using_non_existent_field(self, async_client: AsyncClient, owner_auth_header: dict):
        dataset = await DatasetFactory.create()

        response = await async_client.post(
            self.url(dataset.id),
            headers=owner_auth_header,
            json={"query": {"text": {"q": "text", "field": "non-existent"}}},
        )

        assert response.status_code == 422

    async def test_with_vector_query_using_record_without_vector(
        self, async_client: AsyncClient, owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()

        await TextFieldFactory.create(name="input", dataset=dataset)
        vector_settings = await VectorSettingsFactory.create(name="vector", dimensions=3, dataset=dataset)

        record = await RecordFactory.create(dataset=dataset)
        record_without_vector = await RecordFactory.create(dataset=dataset)

        await VectorFactory.create(value=[1.0, 2.0, 3.0], vector_settings=vector_settings, record=record)

        response = await async_client.post(
            self.url(dataset.id),
            headers=owner_auth_header,
            json={
                "query": {
                    "vector": {
                        "name": vector_settings.name,
                        "record_id": str(record_without_vector.id),
                    },
                },
            },
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": f"Record `{record_without_vector.id}` does not have a vector for vector settings `{vector_settings.name}`"
        }

    async def test_with_filter(
        self, async_client: AsyncClient, mock_search_engine: SearchEngine, owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()

        question = await RatingQuestionFactory.create(dataset=dataset)

        record_a = await RecordFactory.create(dataset=dataset)
        record_b = await RecordFactory.create(dataset=dataset)

        mock_search_engine.search.return_value = SearchResponses(
            items=[
                SearchResponseItem(record_id=record_a.id, score=1.0),
                SearchResponseItem(record_id=record_b.id, score=0.5),
            ],
            total=2,
        )

        response = await async_client.post(
            self.url(dataset.id),
            headers=owner_auth_header,
            json={
                "filters": {
                    "and": [
                        {
                            "type": "terms",
                            "scope": {"entity": "response", "question": question.name},
                            "values": ["value-a"],
                        },
                        {
                            "type": "range",
                            "scope": {"entity": "suggestion", "question": question.name, "property": "score"},
                            "ge": 0.5,
                        },
                    ],
                },
            },
        )

        assert response.status_code == 200

        mock_search_engine.search.assert_called_once_with(
            dataset=dataset,
            filter=AndFilter(
                filters=[
                    TermsFilter(scope=ResponseFilterScope(question=question.name), values=["value-a"]),
                    RangeFilter(scope=SuggestionFilterScope(question=question.name, property="score"), ge=0.5),
                ]
            ),
            metadata_filters=[],
            offset=0,
            limit=50,
            query=None,
            sort_by=None,
            user_response_status_filter=None,
        )

    async def test_with_sort(
        self, async_client: AsyncClient, mock_search_engine: SearchEngine, owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()

        question = await RatingQuestionFactory.create(dataset=dataset)

        record_a = await RecordFactory.create(dataset=dataset)
        record_b = await RecordFactory.create(dataset=dataset)

        mock_search_engine.search.return_value = SearchResponses(
            items=[
                SearchResponseItem(record_id=record_a.id, score=1.0),
                SearchResponseItem(record_id=record_b.id, score=0.5),
            ],
            total=2,
        )

        response = await async_client.post(
            self.url(dataset.id),
            headers=owner_auth_header,
            json={
                "sort": [
                    {
                        "scope": {"entity": "response", "question": question.name},
                        "order": "asc",
                    },
                    {
                        "scope": {"entity": "suggestion", "question": question.name, "property": "score"},
                        "order": "desc",
                    },
                ]
            },
        )

        assert response.status_code == 200

        mock_search_engine.search.assert_called_once_with(
            dataset=dataset,
            sort=[
                Order(scope=ResponseFilterScope(question=question.name), order=SortOrder.asc),
                Order(scope=SuggestionFilterScope(question=question.name, property="score"), order=SortOrder.desc),
            ],
            metadata_filters=[],
            offset=0,
            limit=50,
            query=None,
            sort_by=None,
            user_response_status_filter=None,
        )

    async def test_with_invalid_filter(self, async_client: AsyncClient, owner_auth_header: dict):
        dataset = await DatasetFactory.create()

        response = await async_client.post(
            self.url(dataset.id),
            headers=owner_auth_header,
            json={
                "filters": {
                    "and": [
                        {
                            "type": "terms",
                            "scope": {"entity": "response", "question": "non-existent"},
                            "values": ["value-a"],
                        }
                    ],
                },
            },
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": f"Question with name `non-existent` not found for dataset with id `{dataset.id}`"
        }

    async def test_with_invalid_sort(self, async_client: AsyncClient, owner_auth_header: dict):
        dataset = await DatasetFactory.create()

        response = await async_client.post(
            self.url(dataset.id),
            headers=owner_auth_header,
            json={
                "query": {
                    "text": {"q": "text", "field": "non-existent"},
                },
                "sort": [{"scope": {"entity": "response", "question": "non-existent"}, "order": "asc"}],
            },
        )

        assert response.status_code == 422
        assert response.json() == {
            "detail": f"Question with name `non-existent` not found for dataset with id `{dataset.id}`"
        }



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging

import pydantic
from fastapi.exceptions import RequestValidationError

from argilla_server.errors.base_errors import BadRequestError, GenericServerError, ServerError, ValidationError

_LOGGER = logging.getLogger("argilla")


def exception_to_argilla_error(error: Exception) -> ServerError:
    if isinstance(error, ServerError):
        return error

    _LOGGER.error(error)
    if isinstance(error, RequestValidationError):
        return ValidationError(error)

    if isinstance(error, AssertionError):
        return BadRequestError(str(error))

    # TODO: here we can extend/specify more error adapters
    return GenericServerError(error=error)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
This module configures the api routes under /api prefix, and
set the required security dependencies if api security is enabled
"""

from fastapi import APIRouter, HTTPException, Request

from argilla_server.apis.v0.handlers import (
    authentication,
    datasets,
    info,
    metrics,
    records,
    records_search,
    records_update,
    text2text,
    text_classification,
    token_classification,
    users,
    workspaces,
)
from argilla_server.apis.v1.handlers import datasets as datasets_v1
from argilla_server.apis.v1.handlers import fields as fields_v1
from argilla_server.apis.v1.handlers import metadata_properties as metadata_properties_v1
from argilla_server.apis.v1.handlers import oauth2 as oauth2_v1
from argilla_server.apis.v1.handlers import questions as questions_v1
from argilla_server.apis.v1.handlers import records as records_v1
from argilla_server.apis.v1.handlers import responses as responses_v1
from argilla_server.apis.v1.handlers import suggestions as suggestions_v1
from argilla_server.apis.v1.handlers import users as users_v1
from argilla_server.apis.v1.handlers import vectors_settings as vectors_settings_v1
from argilla_server.apis.v1.handlers import workspaces as workspaces_v1
from argilla_server.errors.base_errors import __ALL__

api_router = APIRouter(responses={error.HTTP_STATUS: error.api_documentation() for error in __ALL__})


dependencies = []

for router in [
    authentication.router,
    users.router,
    workspaces.router,
    datasets.router,
    info.router,
    metrics.router,
    records.router,
    records_search.router,
    records_update.router,
    text_classification.router,
    token_classification.router,
    text2text.router,
]:
    api_router.include_router(router, dependencies=dependencies)

# API v1
api_router.include_router(datasets_v1.router, prefix="/v1")
api_router.include_router(fields_v1.router, prefix="/v1")
api_router.include_router(questions_v1.router, prefix="/v1")
api_router.include_router(metadata_properties_v1.router, prefix="/v1")
api_router.include_router(records_v1.router, prefix="/v1")
api_router.include_router(responses_v1.router, prefix="/v1")
api_router.include_router(suggestions_v1.router, prefix="/v1")
api_router.include_router(users_v1.router, prefix="/v1")
api_router.include_router(vectors_settings_v1.router, prefix="/v1")
api_router.include_router(workspaces_v1.router, prefix="/v1")
api_router.include_router(oauth2_v1.router, prefix="/v1")


@api_router.route("/{_:path}", methods=["GET", "POST", "PUT", "DELETE", "PATCH"], include_in_schema=False)
def endpoint_not_found_controller(request: Request):
    raise HTTPException(status_code=404, detail=f"Endpoint {request.url.path!r} not found")



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Annotated, Any, Dict, List, Literal, Optional, Union
from uuid import UUID

import fastapi
from typing_extensions import Annotated

from argilla_server.enums import RecordInclude, RecordSortField, SimilarityOrder, SortOrder
from argilla_server.pydantic_v1 import BaseModel, Field, root_validator, validator
from argilla_server.pydantic_v1.utils import GetterDict
from argilla_server.schemas.base import UpdateSchema
from argilla_server.schemas.v1.metadata_properties import MetadataPropertyName
from argilla_server.schemas.v1.responses import Response, ResponseFilterScope, UserResponseCreate
from argilla_server.schemas.v1.suggestions import Suggestion, SuggestionCreate, SuggestionFilterScope
from argilla_server.search_engine import TextQuery

RECORDS_CREATE_MIN_ITEMS = 1
RECORDS_CREATE_MAX_ITEMS = 1000

RECORDS_UPDATE_MIN_ITEMS = 1
RECORDS_UPDATE_MAX_ITEMS = 1000

FILTERS_AND_MIN_ITEMS = 1
FILTERS_AND_MAX_ITEMS = 50

TERMS_FILTER_VALUES_MIN_ITEMS = 1
TERMS_FILTER_VALUES_MAX_ITEMS = 250

SEARCH_RECORDS_QUERY_SORT_MIN_ITEMS = 1
SEARCH_RECORDS_QUERY_SORT_MAX_ITEMS = 10


class RecordGetterDict(GetterDict):
    def get(self, key: str, default: Any) -> Any:
        if key == "metadata":
            return getattr(self._obj, "metadata_", None)

        if key == "responses" and not self._obj.is_relationship_loaded("responses"):
            return default

        if key == "suggestions" and not self._obj.is_relationship_loaded("suggestions"):
            return default

        if key == "vectors":
            if self._obj.is_relationship_loaded("vectors"):
                return {vector.vector_settings.name: vector.value for vector in self._obj.vectors}
            else:
                return default

        return super().get(key, default)


class Record(BaseModel):
    id: UUID
    fields: Dict[str, Any]
    metadata: Optional[Dict[str, Any]]
    external_id: Optional[str]
    # TODO: move `responses` to `response` since contextualized endpoint will contains only the user response
    # response: Optional[Response]
    responses: Optional[List[Response]]
    suggestions: Optional[List[Suggestion]]
    vectors: Optional[Dict[str, List[float]]]
    dataset_id: UUID
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True
        getter_dict = RecordGetterDict


class RecordCreate(BaseModel):
    fields: Dict[str, Any]
    metadata: Optional[Dict[str, Any]]
    external_id: Optional[str]
    responses: Optional[List[UserResponseCreate]]
    suggestions: Optional[List[SuggestionCreate]]
    vectors: Optional[Dict[str, List[float]]]

    @validator("responses")
    def check_user_id_is_unique(cls, values: Optional[List[UserResponseCreate]]) -> Optional[List[UserResponseCreate]]:
        if values is None:
            return values

        user_ids = []
        for value in values:
            if value.user_id in user_ids:
                raise ValueError(f"'responses' contains several responses for the same user_id={str(value.user_id)!r}")
            user_ids.append(value.user_id)

        return values

    @validator("metadata", pre=True)
    @classmethod
    def prevent_nan_values(cls, metadata: Optional[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
        if metadata is None:
            return metadata

        for k, v in metadata.items():
            if v != v:
                raise ValueError(f"NaN is not allowed as metadata value, found NaN for key {k!r}")

        return metadata


class RecordUpdate(UpdateSchema):
    metadata_: Optional[Dict[str, Any]] = Field(None, alias="metadata")
    suggestions: Optional[List[SuggestionCreate]] = None
    vectors: Optional[Dict[str, List[float]]]

    @validator("metadata_", pre=True)
    @classmethod
    def prevent_nan_values(cls, metadata: Optional[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
        if metadata is None:
            return metadata

        for k, v in metadata.items():
            if v != v:
                raise ValueError(f"NaN is not allowed as metadata value, found NaN for key {k!r}")

        return {k: v for k, v in metadata.items() if v == v}  # By definition, NaN != NaN


class RecordUpdateWithId(RecordUpdate):
    id: UUID


class RecordIncludeParam(BaseModel):
    relationships: Optional[List[RecordInclude]] = Field(None, alias="keys")
    vectors: Optional[List[str]] = Field(None, alias="vectors")

    @root_validator(skip_on_failure=True)
    def check(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        relationships = values.get("relationships")
        if not relationships:
            return values

        vectors = values.get("vectors")
        if vectors is not None and len(vectors) > 0 and RecordInclude.vectors in relationships:
            # TODO: once we have a exception handler for ValueError in v1, remove HTTPException
            # raise ValueError("Cannot include both 'vectors' and 'relationships' in the same request")
            raise ValueError(
                "'include' query param cannot have both 'vectors' and 'vectors:vector_settings_name_1,vectors_settings_name_2,...'",
            )

        return values

    @property
    def with_responses(self) -> bool:
        return self._has_relationships and RecordInclude.responses in self.relationships

    @property
    def with_suggestions(self) -> bool:
        return self._has_relationships and RecordInclude.suggestions in self.relationships

    @property
    def with_all_vectors(self) -> bool:
        return self._has_relationships and not self.vectors and RecordInclude.vectors in self.relationships

    @property
    def with_some_vector(self) -> bool:
        return self.vectors is not None and len(self.vectors) > 0

    @property
    def _has_relationships(self):
        return self.relationships is not None


class RecordFilterScope(BaseModel):
    entity: Literal["record"]
    property: Union[Literal[RecordSortField.inserted_at], Literal[RecordSortField.updated_at]]


class Records(BaseModel):
    items: List[Record]
    # TODO(@frascuchon): Make it required once fetch records without metadata filter computes also the total
    total: Optional[int] = None


class RecordsCreate(BaseModel):
    items: List[RecordCreate] = Field(..., min_items=RECORDS_CREATE_MIN_ITEMS, max_items=RECORDS_CREATE_MAX_ITEMS)


class RecordsUpdate(BaseModel):
    # TODO: review this definition and align to create model
    items: List[RecordUpdateWithId] = Field(..., min_items=RECORDS_UPDATE_MIN_ITEMS, max_items=RECORDS_UPDATE_MAX_ITEMS)


class MetadataParsedQueryParam:
    def __init__(self, string: str):
        k, *v = string.split(":", maxsplit=1)

        self.name: str = k
        self.value: str = "".join(v).strip()


class MetadataQueryParams(BaseModel):
    metadata: List[str] = Field(fastapi.Query([], pattern=r"^(?=.*[a-z0-9])[a-z0-9_-]+:(.+(,(.+))*)$"))

    @property
    def metadata_parsed(self) -> List[MetadataParsedQueryParam]:
        # TODO: Validate metadata fields names from query params
        return [MetadataParsedQueryParam(q) for q in self.metadata]


class VectorQuery(BaseModel):
    name: str
    record_id: Optional[UUID] = None
    value: Optional[List[float]] = None
    order: SimilarityOrder = SimilarityOrder.most_similar

    @root_validator(skip_on_failure=True)
    def check_required(cls, values: dict) -> dict:
        """Check that either 'record_id' or 'value' is provided"""
        record_id = values.get("record_id")
        value = values.get("value")

        if bool(record_id) == bool(value):
            raise ValueError("Either 'record_id' or 'value' must be provided")

        return values


class Query(BaseModel):
    text: Optional[TextQuery] = None
    vector: Optional[VectorQuery] = None


class MetadataFilterScope(BaseModel):
    entity: Literal["metadata"]
    metadata_property: MetadataPropertyName


FilterScope = Annotated[
    Union[RecordFilterScope, ResponseFilterScope, SuggestionFilterScope, MetadataFilterScope],
    Field(..., discriminator="entity"),
]


class Order(BaseModel):
    scope: FilterScope
    order: SortOrder


class TermsFilter(BaseModel):
    type: Literal["terms"]
    scope: FilterScope
    values: List[str] = Field(..., min_items=TERMS_FILTER_VALUES_MIN_ITEMS, max_items=TERMS_FILTER_VALUES_MAX_ITEMS)


class RangeFilter(BaseModel):
    type: Literal["range"]
    scope: FilterScope
    ge: Optional[float]
    le: Optional[float]

    @root_validator(skip_on_failure=True)
    def check_ge_and_le(cls, values: dict) -> dict:
        ge, le = values.get("ge"), values.get("le")

        if ge is None and le is None:
            raise ValueError("At least one of 'ge' or 'le' must be provided")

        if ge is not None and le is not None and ge > le:
            raise ValueError("'ge' must have a value less than or equal to 'le'")

        return values


Filter = Annotated[Union[TermsFilter, RangeFilter], Field(..., discriminator="type")]


class Filters(BaseModel):
    and_: List[Filter] = Field(None, alias="and", min_items=FILTERS_AND_MIN_ITEMS, max_items=FILTERS_AND_MAX_ITEMS)


class SearchRecordsQuery(BaseModel):
    query: Optional[Query]
    filters: Optional[Filters]
    sort: Optional[List[Order]] = Field(
        None, min_items=SEARCH_RECORDS_QUERY_SORT_MIN_ITEMS, max_items=SEARCH_RECORDS_QUERY_SORT_MAX_ITEMS
    )


class SearchRecord(BaseModel):
    record: Record
    query_score: Optional[float]


class SearchRecordsResult(BaseModel):
    items: List[SearchRecord]
    total: int = 0



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest
from argilla_server.apis.v0.models.text_classification import (
    TextClassificationAnnotation,
    TextClassificationQuery,
    TextClassificationRecord,
)
from argilla_server.commons.models import PredictionStatus, TaskStatus
from argilla_server.daos.backend.search.query_builder import EsQueryBuilder
from argilla_server.services.tasks.text_classification.model import (
    ClassPrediction,
    ServiceTextClassificationRecord,
)

from tests.pydantic_v1 import ValidationError


def test_flatten_metadata():
    data = {
        "inputs": {"text": "bogh"},
        "metadata": {"mail": {"subject": "The mail subject", "body": "This is a large text body"}},
    }
    record = ServiceTextClassificationRecord.parse_obj(data)
    assert list(record.metadata.keys()) == ["mail.subject", "mail.body"]


def test_metadata_with_object_list():
    data = {
        "inputs": {"text": "bogh"},
        "metadata": {
            "mails": [
                {"subject": "Mail One", "body": "This is a large text body"},
                {"subject": "Mail Two", "body": "This is a large text body"},
            ]
        },
    }
    record = ServiceTextClassificationRecord.parse_obj(data)
    assert list(record.metadata.keys()) == ["mails"]


def test_model_dict():
    record = TextClassificationRecord.parse_obj(
        {
            "id": 1,
            "inputs": {"text": "This is a text"},
            "annotation": {
                "agent": "test",
                "labels": [{"class": "A"}, {"class": "B"}],
            },
            "multi_label": True,
        }
    )

    assert record.dict(exclude_none=True) == {
        "annotation": {
            "agent": "test",
            "labels": [
                {"class_label": "A", "score": 1.0},
                {"class_label": "B", "score": 1.0},
            ],
        },
        "annotations": {
            "test": {
                "labels": [
                    {"class_label": "A", "score": 1.0},
                    {"class_label": "B", "score": 1.0},
                ]
            }
        },
        "id": 1,
        "inputs": {"text": "This is a text"},
        "metrics": {},
        "multi_label": True,
        "status": "Default",
    }


def test_model_with_annotations():
    record = TextClassificationRecord.parse_obj(
        {
            "annotations": {
                "test": {
                    "labels": [
                        {"class_label": "A", "score": 1.0},
                        {"class_label": "B", "score": 1.0},
                    ]
                }
            },
            "id": 1,
            "inputs": {"text": "This is a text"},
            "multi_label": True,
            "status": "Default",
        }
    )

    assert record.dict(exclude_none=True) == {
        "annotation": {
            "agent": "test",
            "labels": [
                {"class_label": "A", "score": 1.0},
                {"class_label": "B", "score": 1.0},
            ],
        },
        "annotations": {
            "test": {
                "labels": [
                    {"class_label": "A", "score": 1.0},
                    {"class_label": "B", "score": 1.0},
                ]
            }
        },
        "id": 1,
        "inputs": {"text": "This is a text"},
        "metrics": {},
        "multi_label": True,
        "status": "Default",
    }


def test_single_label_with_multiple_annotation():
    with pytest.raises(
        ValidationError,
        match="Single label record must include only one annotation label",
    ):
        ServiceTextClassificationRecord.parse_obj(
            {
                "inputs": {"text": "This is a text"},
                "annotation": {
                    "agent": "test",
                    "labels": [{"class": "A"}, {"class": "B"}],
                },
                "multi_label": False,
            }
        )


def test_too_long_metadata():
    record = ServiceTextClassificationRecord.parse_obj(
        {
            "inputs": {"text": "bogh"},
            "metadata": {"too_long": "a" * 1000},
        }
    )

    assert len(record.metadata["too_long"]) == 128


def test_too_long_label():
    with pytest.raises(ValidationError, match="exceeds max length"):
        ServiceTextClassificationRecord.parse_obj(
            {
                "inputs": {"text": "bogh"},
                "prediction": {
                    "agent": "test",
                    "labels": [{"class": "a" * 1000}],
                },
            }
        )


def test_score_integrity():
    data = {
        "multi_label": False,
        "inputs": {"data": "My cool data"},
        "prediction": {
            "agent": "test",
            "labels": [
                {"class": "A", "score": 0.3},
                {"class": "B", "score": 0.9},
            ],
        },
    }

    try:
        ServiceTextClassificationRecord.parse_obj(data)
    except ValidationError as e:
        assert "Wrong score distributions" in e.json()

    data["multi_label"] = True
    record = ServiceTextClassificationRecord.parse_obj(data)
    assert record is not None

    data["multi_label"] = False
    data["prediction"]["labels"] = [
        {"class": "B", "score": 0.9},
    ]
    record = ServiceTextClassificationRecord.parse_obj(data)
    assert record is not None

    data["prediction"]["labels"] = [
        {"class": "B", "score": 0.10000000012},
        {"class": "B", "score": 0.90000000002},
    ]
    record = ServiceTextClassificationRecord.parse_obj(data)
    assert record is not None


def test_prediction_ok_cases():
    data = {
        "multi_label": True,
        "inputs": {"data": "My cool data"},
        "prediction": {
            "agent": "test",
            "labels": [
                {"class": "A", "score": 0.3},
                {"class": "B", "score": 0.9},
            ],
        },
    }

    record = ServiceTextClassificationRecord(**data)
    assert record.predicted is None
    record.annotation = TextClassificationAnnotation(
        **{
            "agent": "test",
            "labels": [
                {"class": "A", "score": 1},
                {"class": "B", "score": 1},
            ],
        },
    )
    assert record.predicted == PredictionStatus.KO

    record.prediction = TextClassificationAnnotation(
        **{
            "agent": "test",
            "labels": [
                {"class": "A", "score": 0.9},
                {"class": "B", "score": 0.9},
            ],
        },
    )
    assert record.predicted == PredictionStatus.OK

    record.prediction = None
    assert record.predicted is None


def test_score_ranges():
    with pytest.raises(ValidationError, match="less than or equal to 1.0"):
        ClassPrediction(class_label="BB", score=100)

    with pytest.raises(ValidationError, match="greater than or equal to 0.0"):
        ClassPrediction(class_label="BB", score=-100)


def test_predicted_as_with_no_labels():
    data = {
        "inputs": {"text": "The input text"},
        "prediction": {"agent": "test", "labels": []},
    }
    record = ServiceTextClassificationRecord(**data)
    assert record.predicted_as == []


def test_created_record_with_default_status():
    data = {
        "inputs": {"data": "My cool data"},
    }

    record = ServiceTextClassificationRecord.parse_obj(data)
    assert record.status == TaskStatus.default


def test_predicted_ok_for_multilabel_unordered():
    record = ServiceTextClassificationRecord(
        inputs={"text": "The text"},
        prediction=TextClassificationAnnotation(
            agent="test",
            labels=[
                ClassPrediction(class_label="B"),
                ClassPrediction(class_label="C", score=0.3),
                ClassPrediction(class_label="A"),
            ],
        ),
        annotation=TextClassificationAnnotation(
            agent="test",
            labels=[ClassPrediction(class_label="A"), ClassPrediction(class_label="B")],
        ),
        multi_label=True,
    )

    assert record.predicted == PredictionStatus.OK


@pytest.mark.parametrize(
    "annotation",
    [
        TextClassificationAnnotation(
            agent="test_ok",
            labels=[],
        ),
        None,
    ],
)
def test_validate_without_labels_for_single_label(annotation):
    with pytest.raises(
        ValidationError,
        match="Annotation must include some label for validated records",
    ):
        ServiceTextClassificationRecord(
            inputs={"text": "The text"},
            status=TaskStatus.validated,
            prediction=TextClassificationAnnotation(
                agent="test",
                labels=[
                    ClassPrediction(class_label="C", score=0.3),
                ],
            ),
            annotation=annotation,
        )


def test_query_with_uncovered_by_rules():
    query = TextClassificationQuery(uncovered_by_rules=["query", "other*"])

    assert EsQueryBuilder._to_es_query(query) == {
        "bool": {
            "must": {"match_all": {}},
            "must_not": {
                "bool": {
                    "minimum_should_match": 1,
                    "should": [
                        {
                            "bool": {
                                "must": {
                                    "query_string": {
                                        "default_field": "text",
                                        "default_operator": "AND",
                                        "query": "query",
                                    }
                                }
                            }
                        },
                        {
                            "bool": {
                                "must": {
                                    "query_string": {
                                        "default_field": "text",
                                        "default_operator": "AND",
                                        "query": "other*",
                                    }
                                }
                            }
                        },
                    ],
                }
            },
        }
    }


def test_empty_labels_for_no_multilabel():
    with pytest.raises(
        ValidationError,
        match="Single label record must include only one annotation label",
    ):
        ServiceTextClassificationRecord(
            inputs={"text": "The input text"},
            annotation=TextClassificationAnnotation(agent="ann.", labels=[]),
        )

    record = ServiceTextClassificationRecord(
        inputs={"text": "The input text"},
        prediction=TextClassificationAnnotation(agent="ann.", labels=[]),
        annotation=TextClassificationAnnotation(agent="ann.", labels=[ClassPrediction(class_label="B")]),
    )
    assert record.predicted == PredictionStatus.KO


def test_annotated_without_labels_for_multilabel():
    record = ServiceTextClassificationRecord(
        inputs={"text": "The input text"},
        multi_label=True,
        prediction=TextClassificationAnnotation(agent="pred.", labels=[]),
        annotation=TextClassificationAnnotation(agent="ann.", labels=[]),
    )

    assert record.predicted == PredictionStatus.OK


def test_using_predictions_dict():
    record = ServiceTextClassificationRecord(
        inputs={"text": "this is a text"},
        predictions={
            "carl": TextClassificationAnnotation(agent="wat at", labels=[ClassPrediction(class_label="YES")]),
            "BOB": TextClassificationAnnotation(agent="wot wot", labels=[ClassPrediction(class_label="NO")]),
        },
    )

    assert record.prediction.dict() == {
        "agent": "carl",
        "labels": [{"class_label": "YES", "score": 1.0}],
    }
    assert record.predictions == {
        "BOB": TextClassificationAnnotation(labels=[ClassPrediction(class_label="NO")]),
        "carl": TextClassificationAnnotation(labels=[ClassPrediction(class_label="YES")]),
    }


def test_with_no_agent_at_all():
    with pytest.raises(ValidationError):
        ServiceTextClassificationRecord(
            inputs={"text": "this is a text"},
            prediction=TextClassificationAnnotation(labels=[ClassPrediction(class_label="YES")]),
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING
from uuid import UUID, uuid4

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import User, Workspace, WorkspaceUser
from fastapi.testclient import TestClient
from sqlalchemy import func, select

from tests.factories import (
    AdminFactory,
    AnnotatorFactory,
    UserFactory,
    WorkspaceFactory,
    WorkspaceUserFactory,
)

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession


@pytest.mark.asyncio
async def test_list_workspaces(async_client: "AsyncClient", owner_auth_header):
    await WorkspaceFactory.create(name="workspace-a")
    await WorkspaceFactory.create(name="workspace-b")

    response = await async_client.get("/api/workspaces", headers=owner_auth_header)

    assert response.status_code == 200

    response_body = response.json()
    assert list(map(lambda ws: ws["name"], response_body)) == ["workspace-a", "workspace-b"]


@pytest.mark.asyncio
async def test_list_workspaces_without_authentication(async_client: "AsyncClient"):
    response = await async_client.get("/api/workspaces")

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_list_workspaces_as_admin(async_client: "AsyncClient"):
    admin = await AdminFactory.create()

    response = await async_client.get("/api/workspaces", headers={API_KEY_HEADER_NAME: admin.api_key})

    assert response.status_code == 200


@pytest.mark.asyncio
async def test_list_workspaces_as_annotator(async_client: "AsyncClient"):
    annotator = await AnnotatorFactory.create()

    response = await async_client.get("/api/workspaces", headers={API_KEY_HEADER_NAME: annotator.api_key})

    assert response.status_code == 200


@pytest.mark.asyncio
async def test_create_workspace(async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict):
    response = await async_client.post("/api/workspaces", headers=owner_auth_header, json={"name": "workspace"})

    assert response.status_code == 200
    assert (await db.execute(select(func.count(Workspace.id)))).scalar() == 1

    response_body = response.json()
    assert response_body["name"] == "workspace"
    assert await db.get(Workspace, UUID(response_body["id"]))


@pytest.mark.asyncio
async def test_create_workspace_without_authentication(async_client: "AsyncClient", db: "AsyncSession"):
    response = await async_client.post("/api/workspaces", json={"name": "workspace"})

    assert response.status_code == 401
    assert (await db.execute(select(func.count(Workspace.id)))).scalar() == 0


@pytest.mark.asyncio
async def test_create_workspace_as_admin(async_client: "AsyncClient", db: "AsyncSession"):
    admin = await AdminFactory.create()

    response = await async_client.post(
        "/api/workspaces", headers={API_KEY_HEADER_NAME: admin.api_key}, json={"name": "workspaces"}
    )

    assert response.status_code == 403
    assert (await db.execute(select(func.count(Workspace.id)))).scalar() == 0


@pytest.mark.asyncio
async def test_create_workspace_as_annotator(async_client: "AsyncClient", db: "AsyncSession"):
    annotator = await AnnotatorFactory.create()

    response = await async_client.post(
        "/api/workspaces", headers={API_KEY_HEADER_NAME: annotator.api_key}, json={"name": "workspaces"}
    )

    assert response.status_code == 403
    assert (await db.execute(select(func.count(Workspace.id)))).scalar() == 0


@pytest.mark.asyncio
async def test_create_workspace_with_existent_name(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    await WorkspaceFactory.create(name="workspace")

    response = await async_client.post("/api/workspaces", headers=owner_auth_header, json={"name": "workspace"})

    assert response.status_code == 409
    assert (await db.execute(select(func.count(Workspace.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_create_workspace_with_invalid_min_length_name(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    response = await async_client.post("/api/workspaces", headers=owner_auth_header, json={"name": ""})

    assert response.status_code == 422
    assert (await db.execute(select(func.count(Workspace.id)))).scalar() == 0


@pytest.mark.asyncio
async def test_create_workspace_with_invalid_name(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    response = await async_client.post("/api/workspaces", headers=owner_auth_header, json={"name": "invalid name"})

    assert response.status_code == 422
    assert (await db.execute(select(func.count(Workspace.id)))).scalar() == 0


@pytest.mark.skip(reason="we are not allowing deletion of workspaces right now")
@pytest.mark.asyncio
async def test_delete_workspace():
    pass


@pytest.mark.skip(reason="we are not allowing deletion of workspaces right now")
@pytest.mark.asyncio
async def test_delete_workspace_without_authentication():
    pass


@pytest.mark.skip(reason="we are not allowing deletion of workspaces right now")
@pytest.mark.asyncio
async def test_delete_workspace_as_annotator():
    pass


@pytest.mark.skip(reason="we are not allowing deletion of workspaces right now")
@pytest.mark.asyncio
async def test_delete_workspace_with_nonexistent_workspace_id():
    pass


@pytest.mark.asyncio
async def test_list_workspace_users(async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict):
    workspace_a = await WorkspaceFactory.create()
    user_a = await UserFactory.create(username="username-a")
    user_b = await UserFactory.create(username="username-b")
    user_c = await UserFactory.create(username="username-c")
    await WorkspaceUserFactory.create(workspace_id=workspace_a.id, user_id=user_a.id)
    await WorkspaceUserFactory.create(workspace_id=workspace_a.id, user_id=user_b.id)
    await WorkspaceUserFactory.create(workspace_id=workspace_a.id, user_id=user_c.id)

    user_d = await UserFactory.create(username="username-d")
    user_e = await UserFactory.create(username="username-e")
    await WorkspaceFactory.create(users=[user_d, user_e])

    response = await async_client.get(f"/api/workspaces/{workspace_a.id}/users", headers=owner_auth_header)

    assert response.status_code == 200
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 5

    response_body = response.json()
    assert list(map(lambda u: u["username"], response_body)) == ["username-a", "username-b", "username-c"]


@pytest.mark.asyncio
async def test_list_workspace_users_without_authentication(async_client: "AsyncClient"):
    workspace = await WorkspaceFactory.create()

    response = await async_client.get(f"/api/workspaces/{workspace.id}/users")

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_list_workspace_users_as_admin(async_client: "AsyncClient", db: "AsyncSession"):
    admin = await AdminFactory.create()
    workspace = await WorkspaceFactory.create()
    await WorkspaceUserFactory.create(workspace_id=workspace.id, user_id=admin.id)

    user_a = await UserFactory.create(username="username-a")
    user_b = await UserFactory.create(username="username-b")
    user_c = await UserFactory.create(username="username-c")
    await WorkspaceUserFactory.create(workspace_id=workspace.id, user_id=user_a.id)
    await WorkspaceUserFactory.create(workspace_id=workspace.id, user_id=user_b.id)
    await WorkspaceUserFactory.create(workspace_id=workspace.id, user_id=user_c.id)

    response = await async_client.get(
        f"/api/workspaces/{workspace.id}/users", headers={API_KEY_HEADER_NAME: admin.api_key}
    )

    assert response.status_code == 200
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 4

    response_body = response.json()
    assert list(map(lambda u: u["username"], response_body)) == [
        admin.username,
        "username-a",
        "username-b",
        "username-c",
    ]


@pytest.mark.asyncio
async def test_list_workspace_users_as_annotator(async_client: "AsyncClient"):
    annotator = await AnnotatorFactory.create()
    workspace = await WorkspaceFactory.create()

    response = await async_client.get(
        f"/api/workspaces/{workspace.id}/users", headers={API_KEY_HEADER_NAME: annotator.api_key}
    )

    assert response.status_code == 403


@pytest.mark.asyncio
async def test_create_workspace_user(
    async_client: "AsyncClient", db: "AsyncSession", owner: "User", owner_auth_header: dict
):
    workspace = await WorkspaceFactory.create()

    response = await async_client.post(f"/api/workspaces/{workspace.id}/users/{owner.id}", headers=owner_auth_header)

    assert response.status_code == 200
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 1
    assert (await db.execute(select(WorkspaceUser).filter_by(workspace_id=workspace.id, user_id=owner.id))).scalar_one()

    response_body = response.json()
    assert response_body["id"] == str(owner.id)
    assert workspace.name in response_body["workspaces"]


@pytest.mark.asyncio
async def test_create_workspace_user_without_authentication(async_client: "AsyncClient", db: "AsyncSession", owner):
    workspace = await WorkspaceFactory.create()

    response = await async_client.post(f"/api/workspaces/{workspace.id}/users/{owner.id}")

    assert response.status_code == 401
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 0


@pytest.mark.asyncio
async def test_create_workspace_user_as_admin(async_client: "AsyncClient", db: "AsyncSession"):
    admin = await AdminFactory.create()
    workspace = await WorkspaceFactory.create()
    user = await UserFactory.create()

    response = await async_client.post(
        f"/api/workspaces/{workspace.id}/users/{user.id}", headers={API_KEY_HEADER_NAME: admin.api_key}
    )

    assert response.status_code == 403
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 0


@pytest.mark.asyncio
async def test_create_workspace_user_as_annotator(async_client: "AsyncClient", db: "AsyncSession"):
    annotator = await AnnotatorFactory.create()
    workspace = await WorkspaceFactory.create()
    user = await UserFactory.create()

    response = await async_client.post(
        f"/api/workspaces/{workspace.id}/users/{user.id}", headers={API_KEY_HEADER_NAME: annotator.api_key}
    )

    assert response.status_code == 403
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 0


@pytest.mark.asyncio
async def test_create_workspace_user_with_nonexistent_workspace_id(
    async_client: "AsyncClient", db: "AsyncSession", owner: "User", owner_auth_header: dict
):
    response = await async_client.post(f"/api/workspaces/{uuid4()}/users/{owner.id}", headers=owner_auth_header)

    assert response.status_code == 404
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 0


@pytest.mark.asyncio
async def test_create_workspace_user_with_nonexistent_user_id(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    workspace = await WorkspaceFactory.create()

    response = await async_client.post(f"/api/workspaces/{workspace.id}/users/{uuid4()}", headers=owner_auth_header)

    assert response.status_code == 404
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 0


@pytest.mark.asyncio
async def test_create_workspace_user_with_existent_workspace_id_and_user_id(
    async_client: "AsyncClient", db: "AsyncSession", owner: "User", owner_auth_header: dict
):
    workspace = await WorkspaceFactory.create()
    await WorkspaceUserFactory.create(workspace_id=workspace.id, user_id=owner.id)

    response = await async_client.post(f"/api/workspaces/{workspace.id}/users/{owner.id}", headers=owner_auth_header)

    assert response.status_code == 409
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_workspace_user(async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict):
    workspace = await WorkspaceFactory.create()
    user = await UserFactory.create()
    workspace_user = await WorkspaceUserFactory.create(workspace_id=workspace.id, user_id=user.id)

    response = await async_client.delete(
        f"/api/workspaces/{workspace_user.workspace_id}/users/{workspace_user.user_id}", headers=owner_auth_header
    )

    assert response.status_code == 200
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 0

    response_body = response.json()
    assert response_body["id"] == str(workspace_user.user_id)


@pytest.mark.asyncio
async def test_delete_workspace_user_without_authentication(async_client: "AsyncClient", db: "AsyncSession"):
    workspace = await WorkspaceFactory.create()
    user = await UserFactory.create()
    workspace_user = await WorkspaceUserFactory.create(workspace_id=workspace.id, user_id=user.id)

    response = await async_client.delete(
        f"/api/workspaces/{workspace_user.workspace_id}/users/{workspace_user.user_id}"
    )

    assert response.status_code == 401
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_workspace_user_as_admin(async_client: "AsyncClient", db: "AsyncSession"):
    admin = await AdminFactory.create()
    workspace = await WorkspaceFactory.create()

    await WorkspaceUserFactory.create(workspace_id=workspace.id, user_id=admin.id)
    user = await UserFactory.create()
    workspace_user = await WorkspaceUserFactory.create(workspace_id=workspace.id, user_id=user.id)

    response = await async_client.delete(
        f"/api/workspaces/{workspace_user.workspace_id}/users/{workspace_user.user_id}",
        headers={API_KEY_HEADER_NAME: admin.api_key},
    )

    assert response.status_code == 200
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 1

    response_body = response.json()
    assert response_body["id"] == str(workspace_user.user_id)


@pytest.mark.asyncio
async def test_delete_workspace_user_as_annotator(async_client: "AsyncClient", db: "AsyncSession"):
    annotator = await AnnotatorFactory.create()
    workspace = await WorkspaceFactory.create()
    user = await UserFactory.create()
    workspace_user = await WorkspaceUserFactory.create(
        workspace_id=workspace.id,
        user_id=user.id,
    )

    response = await async_client.delete(
        f"/api/workspaces/{workspace_user.workspace_id}/users/{workspace_user.user_id}",
        headers={API_KEY_HEADER_NAME: annotator.api_key},
    )

    assert response.status_code == 403
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_workspace_user_with_nonexistent_workspace_id(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    workspace = await WorkspaceFactory.create()
    user = await UserFactory.create()
    workspace_user = await WorkspaceUserFactory.create(workspace_id=workspace.id, user_id=user.id)

    response = await async_client.delete(
        f"/api/workspaces/{uuid4()}/users/{workspace_user.user_id}", headers=owner_auth_header
    )

    assert response.status_code == 404
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_workspace_user_with_nonexistent_user_id(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    workspace = await WorkspaceFactory.create()
    user = await UserFactory.create()
    workspace_user = await WorkspaceUserFactory.create(workspace_id=workspace.id, user_id=user.id)

    response = await async_client.delete(
        f"/api/workspaces/{workspace_user.workspace_id}/users/{uuid4()}", headers=owner_auth_header
    )

    assert response.status_code == 404
    assert (await db.execute(select(func.count(WorkspaceUser.id)))).scalar() == 1
